import json
import logging
import math
import sys
from ast import parse
from datetime import datetime, timedelta, timezone
from unittest.mock import patch

import pytest
import requests
from jira import JIRAError
import requests_mock

from jf_ingest.config import IssueMetadata
from jf_ingest.utils import (
    RETRY_EXPONENT_BASE,
    RetryLimitExceeded,
    batch_iterable,
    batch_iterable_by_bytes_size,
    format_date_to_jql,
    get_object_bytes_size,
    get_wait_time,
    hash_filename,
    parse_gitlab_api_version,
    retry_for_status,
    normalize_datetime
)

logger = logging.getLogger(__name__)

MOCK_RETRY_AFTER_TIME = 12345


class MockJiraErrorResponseWithRetryAfterAsInt:
    headers = {"Retry-After": MOCK_RETRY_AFTER_TIME}


class MockJiraErrorResponseWithRetryAfterAsStr:
    headers = {"Retry-After": str(MOCK_RETRY_AFTER_TIME)}


class MockJiraErrorResponseWithZeroRetryAfter:
    headers = {"Retry-After": 0}


class MockJiraErrorResponseWithNonIntRetryAfter:
    headers = {"Retry-After": "CAT"}


MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsInt(),
)
MOCKED_504_JIRA_ERROR_WITH_RETRY_AS_INT = JIRAError(
    status_code=504,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsInt(),
)
MOCKED_404_JIRA_ERROR_WITH_RETRY_AS_INT = JIRAError(
    status_code=404,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsInt(),
)
MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_STR = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponseWithRetryAfterAsStr(),
)

MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_ZERO = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error with Retry After set to 0",
    response=MockJiraErrorResponseWithZeroRetryAfter(),
)

MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_NOT_INT = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error with Retry After set to a non int",
    response=MockJiraErrorResponseWithNonIntRetryAfter(),
)


def test_get_default_wait_time():
    # Get Default Wait Times
    for retry_num in range(0, 10):
        assert get_wait_time(e=None, retries=retry_num) == RETRY_EXPONENT_BASE**retry_num


def test_get_wait_times_for_retry_error():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    for retry_num in range(0, 5):
        logger.info(f"Retry number {retry_num}...")
        assert get_wait_time(mock_error, retries=retry_num) == MOCK_RETRY_AFTER_TIME


def test_get_wait_times_for_retry_error_when_retry_is_str():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_STR

    for retry_num in range(0, 5):
        logger.info(f"Retry number {retry_num}...")
        assert get_wait_time(mock_error, retries=retry_num) == MOCK_RETRY_AFTER_TIME


def test_get_wait_times_for_retry_error_zero():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_ZERO

    for retry_num in range(0, 10):
        logger.info(f"Retry number {retry_num}...")
        expected_wait_time = RETRY_EXPONENT_BASE**retry_num
        # Retry times SHOULD NEVER be zero
        assert get_wait_time(mock_error, retries=retry_num) != 0
        assert get_wait_time(mock_error, retries=retry_num) == expected_wait_time


def test_get_wait_times_for_retry_error_non_zero():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR_RETRY_AFTER_IS_NOT_INT

    for retry_num in range(0, 10):
        logger.info(f"Retry number {retry_num}...")
        expected_wait_time = RETRY_EXPONENT_BASE**retry_num
        # Retry times SHOULD NEVER be zero
        assert get_wait_time(mock_error, retries=retry_num) != 0
        assert get_wait_time(mock_error, retries=retry_num) == expected_wait_time


def test_get_wait_time_for_non_retry_error():
    mock_non_retry_exception = Exception()
    retries = 5
    assert get_wait_time(mock_non_retry_exception, retries=retries) == RETRY_EXPONENT_BASE**retries


def test_retry_for_status_timeout():
    def always_raise_429():
        raise MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        with pytest.raises(RetryLimitExceeded):
            retry_for_status(always_raise_429)

        try:
            retry_for_status(always_raise_429)
        except RetryLimitExceeded:
            m.assert_called_with(MOCK_RETRY_AFTER_TIME)


def test_retry_for_status_retry_works_429():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    def raise_429_three_times():
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert retry_for_status(raise_429_three_times) == success_message
        m.assert_called_with(MOCK_RETRY_AFTER_TIME)
        
def test_session_get_with_retry_for_status(requests_mock: requests_mock.Mocker):
    fake_url = 'https://a-fake-url.com'
    retry_count = 10
    requests_mock.get(url=fake_url, status_code=429)
    
    with patch("jf_ingest.utils.time.sleep", return_value=0):
        session = requests.Session()
        f = session.get
        with pytest.raises(RetryLimitExceeded):
            retry_for_status(f, url=fake_url, max_retries_for_retry_for_status=retry_count)
            assert requests_mock.call_count == retry_count
        
def test_session_post_with_retry_for_status(requests_mock: requests_mock.Mocker):
    fake_url = 'https://a-fake-url.com'
    retry_count = 10
    test_body = {'fake': 'data'}
    requests_mock.post(url=fake_url, json=test_body, status_code=429)
    
    with patch("jf_ingest.utils.time.sleep", return_value=0):
        session = requests.Session()
        f = session.post
        with pytest.raises(RetryLimitExceeded):
            retry_for_status(f, url=fake_url, json=test_body, max_retries_for_retry_for_status=retry_count)
            assert requests_mock.call_count == retry_count
            

def test_session_get_with_retry_for_status_for_non_retryable(requests_mock: requests_mock.Mocker):
    fake_url = 'https://a-fake-url.com'
    retry_count = 10
    requests_mock.get(url=fake_url, status_code=401)
    
    with patch("jf_ingest.utils.time.sleep", return_value=0):
        session = requests.Session()
        f = session.get
        with pytest.raises(requests.exceptions.HTTPError):
            retry_for_status(f, url=fake_url, max_retries_for_retry_for_status=retry_count)
            assert requests_mock.call_count == 1


def test_retry_for_status_retry_works_504():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    def raise_504_three_times():
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise MOCKED_504_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert retry_for_status(raise_504_three_times) == success_message
        m.assert_called_with(MOCK_RETRY_AFTER_TIME)


def test_retry_for_status_retry_works_custom_error_code_list():
    def always_raise_429():
        raise MOCKED_429_JIRA_ERROR_WITH_RETRY_AS_INT

    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    some_error_codes = []
    some_error_codes.extend([400, 500])
    some_error_codes.append(404)

    def raise_404_three_times(test_arg: str):
        assert bool(test_arg)
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise MOCKED_404_JIRA_ERROR_WITH_RETRY_AS_INT

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert (
            retry_for_status(
                raise_404_three_times, 'test_input', statuses_to_retry=some_error_codes
            )
            == success_message
        )
        m.assert_called_with(MOCK_RETRY_AFTER_TIME)

    with pytest.raises(JIRAError):
        # Assert that we raise codes that are not included
        retry_for_status(always_raise_429, statuses_to_retry=some_error_codes)


def test_retry_for_status_with_generic_error():
    def raise_non_429():
        raise Exception("Generic Error Testing")

    with pytest.raises(Exception):
        retry_for_status(raise_non_429)


def test_retry_for_status_retry_works_readtimeout():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    def raise_readtimeout_three_times():
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise requests.exceptions.ReadTimeout()

    expected_wait_time = get_wait_time(None, retries=2)
    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert retry_for_status(raise_readtimeout_three_times) == success_message
        m.assert_called_with(expected_wait_time)


def test_retry_for_status_retry_works_custom_exception():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    class CustomException(Exception):
        pass

    def raise_readtimeout_three_times():
        if arg_dict["retry_count"] == 1:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise CustomException()

    expected_wait_time = get_wait_time(None, retries=0)
    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert (
            retry_for_status(raise_readtimeout_three_times, exceptions_to_retry=(CustomException,))
            == success_message
        )
        m.assert_called_with(expected_wait_time)


def test_retry_for_status_with_jira_non_caught_error():
    def raise_401():
        raise JIRAError(text="Generic Error Testing", status_code=401)

    def raise_429():
        raise JIRAError(text="Generic Error Testing", status_code=429)

    with pytest.raises(JIRAError):
        retry_for_status(raise_401)

    with pytest.raises(JIRAError):
        retry_for_status(raise_429(), status_codes_to_retry=[])

def test_retry_for_status_with_generic_error_raise_or_not():
    class NewExceptionType(Exception):
        pass
    def raise_generic_exception():
        raise NewExceptionType("Generic Error Testing")

    with pytest.raises(NewExceptionType), patch('time.sleep', return_value=0):
        retry_for_status(raise_generic_exception, retry_on_any_exception=False)
    
    
    with pytest.raises(RetryLimitExceeded), patch('time.sleep', return_value=0):
        retry_for_status(raise_generic_exception, retry_on_any_exception=True)


def test_batch_iterable():
    total_data_size = 85926
    batch_size = 62
    data = [i for i in range(total_data_size)]

    recompiled_batches = []
    for batch in batch_iterable(data, batch_size=batch_size):
        # Assert we are getting proper batch sizes
        assert len(batch) <= batch_size
        recompiled_batches.extend(batch)

    # Assert we got ALL the data
    assert len(recompiled_batches) == total_data_size

    # Assert that data is returned matching and in order
    for i in range(total_data_size):
        assert data[i] == recompiled_batches[i]


def test_batch_iterable_by_bytes_size():
    number_of_chars = 41234
    data = ['a' for i in range(number_of_chars)]
    total_data_size_in_bytes = sum(sys.getsizeof(i) for i in data)

    # NOTE: This values should be greater than 50
    batch_bytes_size = 500

    batch_numbers = 0
    recompiled_batches = []
    for batch in batch_iterable_by_bytes_size(data, batch_byte_size=batch_bytes_size):
        # Assert we are getting proper batch sizes
        assert sys.getsizeof(batch) > 0
        assert sys.getsizeof(batch) <= batch_bytes_size
        batch_numbers += 1
        recompiled_batches.extend(batch)

    # Assert we got ALL the data
    assert len(recompiled_batches) == len(data)
    assert batch_numbers == math.ceil(total_data_size_in_bytes / batch_bytes_size)


def test_get_object_bytes_size():
    test_object = {'a': 1, 'b': 2, 'c': 3}

    object_size = get_object_bytes_size(test_object)

    manually_derived_size = 0
    for key in test_object.keys():
        manually_derived_size += sys.getsizeof(key)
        manually_derived_size += sys.getsizeof(test_object[key])
    assert manually_derived_size == object_size


def test_get_object_bytes_size_nested_object():
    test_object = {'a': 1, 'b': 2, 'c': {'d': 4}}

    object_size = get_object_bytes_size(test_object)

    manually_derived_size = 0
    for key in test_object.keys():
        manually_derived_size += sys.getsizeof(key)
        value = test_object[key]
        if type(value) == dict:
            for _key in value.keys():
                manually_derived_size += sys.getsizeof(_key)
                _value = value[_key]
                manually_derived_size += sys.getsizeof(_value)
        else:
            manually_derived_size += sys.getsizeof(value)

    assert manually_derived_size == object_size


def test_format_date_to_jql():
    min_datetime = datetime.min
    assert format_date_to_jql(min_datetime) == '"0001-01-01"'

    _american_revolution_datetime = datetime(1776, 7, 4)
    assert format_date_to_jql(_american_revolution_datetime) == '"1776-07-04"'

    _datetime_pre_unix_epoch_time = datetime(1969, 12, 31)
    assert format_date_to_jql(_datetime_pre_unix_epoch_time) == '"1969-12-31"'

    _datetime_after_unix_epoch_time = datetime(2003, 3, 12)
    assert format_date_to_jql(_datetime_after_unix_epoch_time) == '"2003-03-12"'

    _datetime_after_unix_epoch_time = datetime(2024, 12, 18)
    assert format_date_to_jql(_datetime_after_unix_epoch_time) == '"2024-12-18"'

    _future_datetime = datetime(3333, 1, 4)
    assert format_date_to_jql(_future_datetime) == '"3333-01-04"'


def test_issue_metadata_serializer_single_object():
    # Create an IssueMetadata Object with values we can
    # verify after we do a "round trip" of serializing
    # and deserializing
    id = "1"
    key = "PROJ-1"
    _datetime = datetime(2023, 12, 29)
    epic_link_field_issue_key = "EPIC_LINK_FIELD_ISSUE_KEY"
    parent_field_issue_key = "PARENT_FIELD_ISSUE_KEY"
    issue_metadata = IssueMetadata(
        id=id,
        key=key,
        updated=_datetime,
        epic_link_field_issue_key=epic_link_field_issue_key,
        parent_field_issue_key=parent_field_issue_key,
    )

    # Serialize the object to a string
    serialized_issue_metadata = IssueMetadata.to_json_str(issue_metadata)
    assert type(serialized_issue_metadata) == str

    # Deserialize it back to memory, verify that it is the proper type
    deserialized_issue_metadata = IssueMetadata.from_json(serialized_issue_metadata)
    assert type(deserialized_issue_metadata) == IssueMetadata

    # Assert that all the data is the same after a serialization/deserialization round trip
    assert deserialized_issue_metadata.id == id
    assert deserialized_issue_metadata.key == key
    assert deserialized_issue_metadata.updated == _datetime
    assert deserialized_issue_metadata.epic_link_field_issue_key == epic_link_field_issue_key
    assert deserialized_issue_metadata.parent_field_issue_key == parent_field_issue_key


def test_issue_metadata_serializer_list_of_objects():
    # Generate a list of IssueMetadata to serialize/deserialize
    issue_metadata_list: list[IssueMetadata] = []
    lower_bound_day = 1
    upper_bound_day = 30
    for day in range(lower_bound_day, upper_bound_day):
        id = f"{day}"
        key = f"PROJ-{day}"
        _datetime = datetime(2023, 12, day)
        epic_link_field_issue_key = f"EPIC_LINK_FIELD_ISSUE_KEY_{day}"
        parent_field_issue_key = f"PARENT_FIELD_ISSUE_KEY_{day}"

        issue_metadata_list.append(
            IssueMetadata(
                id=id,
                key=key,
                updated=_datetime,
                epic_link_field_issue_key=epic_link_field_issue_key,
                parent_field_issue_key=parent_field_issue_key,
            )
        )

    # Verify we can serialize it as a list
    serialized_issue_metadata_list = IssueMetadata.to_json_str(issue_metadata_list)
    assert type(serialized_issue_metadata_list) == str

    # Verify that we can deserialize it as a list of the same size and of the same type
    deserialized_issue_metadata_list = IssueMetadata.from_json(serialized_issue_metadata_list)
    assert type(deserialized_issue_metadata_list) == list
    assert len(deserialized_issue_metadata_list) == len(issue_metadata_list)
    for item in deserialized_issue_metadata_list:
        assert type(item) == IssueMetadata

    for i in range(len(issue_metadata_list)):
        issue_metadata_pre = issue_metadata_list[i]
        issue_metadata_post = deserialized_issue_metadata_list[i]

        assert issue_metadata_pre.id == issue_metadata_post.id
        assert issue_metadata_pre.key == issue_metadata_post.key
        assert issue_metadata_pre.updated == issue_metadata_post.updated
        assert (
            issue_metadata_pre.epic_link_field_issue_key
            == issue_metadata_post.epic_link_field_issue_key
        )
        assert (
            issue_metadata_pre.parent_field_issue_key == issue_metadata_post.parent_field_issue_key
        )


def test_hash_files():
    filepath = 'a/path/to/a/file.py'
    hashed_filepath = hash_filename(filepath)
    assert hashed_filepath == '0cc175b9c0f1/d6fe1d0be634/01b6e20344b6/0cc175b9c0f1/02b202aa0962.py'

    # Test when file has additional . in it
    filepath = 'a/path/to.a/file.py'
    hashed_filepath = hash_filename(filepath)
    assert hashed_filepath == '0cc175b9c0f1/d6fe1d0be634/18449aad14e9/02b202aa0962.py'

    # Test when file has a bunch of additional . in it
    filepath = 'a.path/to.a/file.py'
    hashed_filepath = hash_filename(filepath)
    assert hashed_filepath == '8bd1e22aeb9f/18449aad14e9/02b202aa0962.py'

def test_parse_gitlab_api_version():
    versions_to_test = ['17.4.0', '17.4.0-pre', '17.4.0-ee', '17-stuff.4-random.0-things']
    for version_to_test in versions_to_test:
        assert parse_gitlab_api_version(version_to_test) == '17.4.0'

def test_normalize_timezone():
    UTC = timezone.utc
    los_angeles = timezone(timedelta(hours=-8))
    new_york = timezone(timedelta(hours=-5))
    
    base_time = datetime(2024, 1, 1, 18, 0, 0, tzinfo=UTC)
    
    assert normalize_datetime(base_time, UTC) == base_time
    assert normalize_datetime(base_time, los_angeles) == datetime(2024, 1, 1, 10, 0, 0, tzinfo=los_angeles)
    assert normalize_datetime(base_time, new_york) == datetime(2024, 1, 1, 13, 0, 0, tzinfo=new_york)
    
    base_time = datetime(2024, 1, 1, 0, 0, 0, tzinfo=UTC)  # Test date change
    
    assert normalize_datetime(base_time, UTC) == base_time
    assert normalize_datetime(base_time, los_angeles) == datetime(2023, 12, 31, 16, 0, 0, tzinfo=los_angeles)
    assert normalize_datetime(base_time, new_york) == datetime(2023, 12, 31, 19, 0, 0, tzinfo=new_york)
    
    base_time = datetime(1, 1, 1, 1, 1, 1, tzinfo=UTC)  # Test overflow error. If the date is low enough, we don't bother messing with timezone
    
    assert normalize_datetime(base_time, UTC) == base_time
    assert normalize_datetime(base_time, new_york) == base_time
    assert normalize_datetime(base_time, los_angeles) == base_time
    


import re
from functools import wraps
from typing import List

from jf_ingest.name_redactor import NameRedactor

branch_redactor = NameRedactor(preserve_names=['master', 'develop'])
organization_redactor = NameRedactor()
repo_redactor = NameRedactor()


def sanitize_text(text: str) -> str:
    """Helper function for removing everything other than jira keys out of
    text bodies

    Args:
        text (str): A given string to "sanitize"

    Returns:
        str: The inputted text str, but with any Jira Key purged out of it
    """
    # NOTE: This module is used only by git, but we need to clean up Jira
    # keys out of commit messages. That's what this regex is for
    JIRA_KEY_REGEX = re.compile(r'([a-z0-9]+)[-|_|/| ]?(\d+)', re.IGNORECASE)

    if not text:
        return text

    regex_matches: List[str] = JIRA_KEY_REGEX.findall(text)

    return (' ').join(
        {f'{match[0].upper().strip()}-{match[1].upper().strip()}' for match in regex_matches}
    )



import logging
from datetime import datetime
from typing import Any, Dict, Generator, List, Optional, Union

from gitlab.base import RESTObject
from gitlab.v4.objects import Group, Project, ProjectBranch, ProjectCommit
from gitlab.v4.objects import User as GitlabUser

from jf_ingest import logging_helper
from jf_ingest.config import GitConfig, IngestionType
from jf_ingest.jf_git.adapters import GitAdapter
from jf_ingest.jf_git.clients.gitlab import GitlabClient
from jf_ingest.jf_git.standardized_models import (
    PullRequestReviewState,
    StandardizedBranch,
    StandardizedCommit,
    StandardizedFileData,
    StandardizedJFAPIPullRequest,
    StandardizedOrganization,
    StandardizedPullRequest,
    StandardizedPullRequestAuthor,
    StandardizedPullRequestComment,
    StandardizedPullRequestMetadata,
    StandardizedPullRequestReview,
    StandardizedPullRequestReviewAuthor,
    StandardizedRepository,
    StandardizedShortRepository,
    StandardizedTeam,
    StandardizedUser,
)
from jf_ingest.utils import (
    GitLabGidObjectMapping,
    batch_iterable,
    get_attribute,
    get_id_from_gid,
    get_ingestion_type,
    hash_filename,
    parse_gitlab_date,
)

logger = logging.getLogger(__name__)


class GitlabAdapter(GitAdapter):

    def __init__(self, config: GitConfig):
        self.config = config
        self.client = GitlabClient(auth_config=config.git_auth_config)
        self.group_id_to_full_path: Dict[str, str] = {}

    def get_api_scopes(self) -> str:
        """Return the list of API Scopes. This is useful for Validation

        Returns:
            str: A string of API scopes we have, given the adapters credentials
        """
        raise NotImplementedError()

    @staticmethod
    def _get_id_from_gid(gitlab_gid: str, object_name: str) -> str:
        # Convenience wrapper for util function get_id_from_gid
        return get_id_from_gid(gitlab_gid, object_name)

    @staticmethod
    def _standardize_commit_author(api_user: dict) -> StandardizedUser:
        id = api_user.get('id') or api_user.get('email')  # Commit users may not have ids
        name = api_user.get('name')
        login = api_user.get('login') or api_user.get('email')
        email = api_user.get('email')

        return StandardizedUser(
            id=str(id),
            login=str(login),
            name=str(name),
            email=str(email),
        )

    @staticmethod
    def _standardize_gitlab_commit(
        gitlab_commit: ProjectCommit | RESTObject,
        standardized_repo: StandardizedRepository,
        branch_name: str,
    ) -> StandardizedCommit:
        """
        Converts a ProjectCommit from the gitlab client to a standardized jf commit.
        """
        author = GitlabAdapter._standardize_commit_author(
            {
                'name': gitlab_commit.author_name,
                'email': gitlab_commit.author_email,
            }
        )
        return StandardizedCommit(
            hash=gitlab_commit.id,
            author=author,
            url=gitlab_commit.web_url,
            commit_date=parse_gitlab_date(gitlab_commit.committed_date),
            author_date=parse_gitlab_date(gitlab_commit.authored_date),
            message=gitlab_commit.message,
            is_merge=len(gitlab_commit.parent_ids) > 1,
            repo=standardized_repo.short(),  # use short form of repo
            branch_name=branch_name,
        )

    @staticmethod
    def _standardize_gitlab_project(
        gitlab_repo: Project | RESTObject,
        standardized_organization: StandardizedOrganization,
    ) -> StandardizedRepository:
        """
        Converts a Project from the gitlab client to a standardized jf project.
        """
        repo_name = gitlab_repo.name
        default_branch_name: str | None = get_attribute(gitlab_repo, 'default_branch', default=None)
        return StandardizedRepository(
            id=str(gitlab_repo.id),
            name=repo_name,
            full_name=repo_name,
            url=gitlab_repo.web_url,
            default_branch_sha='',
            default_branch_name=default_branch_name,
            organization=standardized_organization,
            is_fork=bool(getattr(gitlab_repo, 'forked_from_project', False)),
            full_path=gitlab_repo.path_with_namespace,
        )

    @staticmethod
    def _standardize_gitlab_branch(
        gitlab_branch: ProjectBranch | RESTObject,
        standardized_repository: StandardizedRepository,
    ) -> StandardizedBranch | None:
        """
        Converts a Branch (ProjectBranch) from the gitlab client to a standardized jf branch.
        """
        if not gitlab_branch:
            return None
        branch_name = get_attribute(gitlab_branch, 'name', default=None)
        return StandardizedBranch(
            name=branch_name,
            sha=gitlab_branch.commit['id'],
            repo_id=standardized_repository.id,
            is_default=gitlab_branch.default,
        )

    @staticmethod
    def _standardize_user(api_user: Union[RESTObject, GitlabUser]) -> StandardizedUser:
        return StandardizedUser(
            id=api_user.id,
            name=api_user.name,
            login=api_user.username,
            email=None,
            url=api_user.web_url,
        )

    @staticmethod
    def _standardize_gitlab_pull_request_gql(
        gitlab_pr: dict,
        standardized_repository: StandardizedRepository,
    ) -> StandardizedPullRequest:
        """
        Converts a gql mergeRequest dict to a standardized jf pull request.
        """
        # Determine closed/merged status and build source/target short repos
        is_closed = bool(gitlab_pr['closedAt'])
        is_merged = bool(gitlab_pr['mergedAt'])
        standardized_head_repository = StandardizedShortRepository(
            id=get_id_from_gid(
                gitlab_pr['sourceProject']['id'], GitLabGidObjectMapping.PROJECT.value
            ),
            name=gitlab_pr['sourceProject']['name'],
            url=gitlab_pr['sourceProject']['webUrl'],
        )
        standardized_base_repository = StandardizedShortRepository(
            id=get_id_from_gid(
                gitlab_pr['targetProject']['id'], GitLabGidObjectMapping.PROJECT.value
            ),
            name=gitlab_pr['targetProject']['name'],
            url=gitlab_pr['targetProject']['webUrl'],
        )

        # Build out standardized commits
        standardized_short_repository: StandardizedShortRepository = standardized_repository.short()
        merge_commit_sha = gitlab_pr['mergeCommitSha']
        branch_name = gitlab_pr['sourceBranch']
        standardized_commits = []
        standardized_merge_commit = None
        for gitlab_commit in gitlab_pr['commits']['nodes']:
            standardized_commit = GitlabAdapter._standardize_gitlab_commit_gql(
                gitlab_commit,
                standardized_short_repository,
                branch_name=branch_name,
                is_merge=gitlab_commit['sha'] == merge_commit_sha,
            )
            standardized_commits.append(standardized_commit)
            if gitlab_commit['sha'] == merge_commit_sha:
                standardized_merge_commit = standardized_commit

        # Build PR comments
        standardized_comments = []
        for gitlab_note in gitlab_pr['notes']['nodes']:
            standardized_comment = GitlabAdapter._standardize_gitlab_pull_request_comment_gql(
                gitlab_note
            )
            standardized_comments.append(standardized_comment)

        # Build PR approvals
        standardized_approvals = []
        # Approvals here are essentially users with more steps
        for gitlab_approval in gitlab_pr['approvedBy']['nodes']:
            standardized_user = GitlabAdapter._standardize_gitlab_user_gql(gitlab_approval)
            standardized_approval = StandardizedPullRequestReview(
                user=standardized_user,
                foreign_id=standardized_user.id if standardized_user else '',
                review_state=PullRequestReviewState.APPROVED.name,
            )
            standardized_approvals.append(standardized_approval)

        # Build file data from diff stats
        standardized_file_data = {}
        for gitlab_file_data in gitlab_pr.get('diffStats', []):
            standardized_file_data[hash_filename(gitlab_file_data['path'])] = StandardizedFileData(
                status='',
                changes=gitlab_file_data['additions'] + gitlab_file_data['deletions'],
                additions=gitlab_file_data['additions'],
                deletions=gitlab_file_data['deletions'],
            )

        base_branch = gitlab_pr['targetBranch']
        head_branch = gitlab_pr['sourceBranch']

        return StandardizedPullRequest(
            id=int(get_id_from_gid(gitlab_pr['id'], GitLabGidObjectMapping.MERGE_REQUEST.value)),
            additions=gitlab_pr['diffStatsSummary']['additions'],
            deletions=gitlab_pr['diffStatsSummary']['deletions'],
            changed_files=gitlab_pr['diffStatsSummary']['fileCount'],
            is_closed=is_closed,
            is_merged=is_merged,
            created_at=parse_gitlab_date(gitlab_pr['createdAt']),
            updated_at=parse_gitlab_date(gitlab_pr['updatedAt']),
            merge_date=parse_gitlab_date(gitlab_pr['mergedAt']),
            closed_date=parse_gitlab_date(gitlab_pr['closedAt']),
            title=gitlab_pr['title'],
            body=gitlab_pr['description'],
            url=gitlab_pr['webUrl'],
            base_branch=base_branch,
            head_branch=head_branch,
            author=GitlabAdapter._standardize_gitlab_user_gql(  # type:ignore[arg-type]
                gitlab_pr['author']
            ),  # This will not be null, but this serialization returns `None` if nothing was passed in.
            merged_by=GitlabAdapter._standardize_gitlab_user_gql(gitlab_pr['mergeUser']),
            commits=standardized_commits,
            merge_commit=standardized_merge_commit,
            comments=standardized_comments,
            approvals=standardized_approvals,
            base_repo=standardized_base_repository,
            head_repo=standardized_head_repository,
            labels=[],  # TODO: Support labels (OJ-7202)
            files=standardized_file_data,
        )

    @staticmethod
    def _standardize_gitlab_user_gql(
        gitlab_user: Optional[dict] = None,
    ) -> Optional[StandardizedUser]:
        """
        Converts a gql user dict to a standardized jf user.
        """
        if not gitlab_user:
            return None

        return StandardizedUser(
            id=get_id_from_gid(gitlab_user['id'], GitLabGidObjectMapping.USER.value),
            login=gitlab_user['username'] or gitlab_user['publicEmail'],
            name=gitlab_user['name'],
            email=gitlab_user['publicEmail'],
            url=gitlab_user['webUrl'],
        )

    @staticmethod
    def _standardize_gitlab_commit_gql(
        gitlab_commit: dict,
        standardized_short_repo: StandardizedShortRepository,
        branch_name: Optional[str] = None,
        is_merge: Optional[bool] = False,
    ) -> StandardizedCommit:
        """
        Converts a gql commit dict to a standardized jf commit.
        """
        return StandardizedCommit(
            hash=gitlab_commit['sha'],
            url=gitlab_commit['webUrl'],
            message=gitlab_commit['message'],
            commit_date=parse_gitlab_date(gitlab_commit['committedDate']),
            author_date=parse_gitlab_date(gitlab_commit['authoredDate']),
            author=GitlabAdapter._standardize_gitlab_user_gql(gitlab_commit['author']),
            repo=standardized_short_repo,
            is_merge=bool(is_merge),
            branch_name=branch_name,
        )

    @staticmethod
    def _standardize_gitlab_pull_request_comment_gql(
        gitlab_pr_comment: dict,
    ) -> StandardizedPullRequestComment:
        """
        Converts a gql comment dict to a standardized jf comment.
        """
        return StandardizedPullRequestComment(
            user=GitlabAdapter._standardize_gitlab_user_gql(gitlab_pr_comment['author']),
            body=gitlab_pr_comment['body'],
            created_at=parse_gitlab_date(gitlab_pr_comment['createdAt']),
            system_generated=gitlab_pr_comment['system'],
        )

    def get_group_full_path_from_id(self, group_id: str) -> str:
        if group_id not in self.group_id_to_full_path:
            _, full_path, _ = self.client.get_organization_name_full_path_and_url(login=group_id)
            self.group_id_to_full_path[group_id] = full_path

        return self.group_id_to_full_path[group_id]

    def get_group_full_path_from_organization(self, org: StandardizedOrganization) -> str:
        return self.get_group_full_path_from_id(org.login)

    def get_organizations(self) -> List[StandardizedOrganization]:
        """Get the list of organizations the adapter has access to

        Returns:
            List[StandardizedOrganization]: A list of standardized organizations within this Git Instance
        """
        orgs: List[StandardizedOrganization] = []
        should_discover: bool = False

        if self.config.discover_organizations:
            if self.client.supports_get_all_organizations:
                should_discover = True
            else:
                logger.info(
                    'Discover new organizations is enabled, but this instance does not support getting all organizations. '
                    'This is because the instance is using the public gitlab.com domain and will return other public '
                    'organizations. Falling back to configured git organizations.'
                )

        if should_discover:
            try:
                # Discover Orgs
                for gql_org in self.client.get_organizations_gql():
                    group_id = get_id_from_gid(
                        gql_org['groupIdStr'], GitLabGidObjectMapping.GROUP.value
                    )
                    full_path = gql_org['fullPath']
                    self.group_id_to_full_path[group_id] = full_path
                    orgs.append(
                        StandardizedOrganization(
                            id=group_id, name=gql_org['name'], login=group_id, url=gql_org['webUrl']
                        )
                    )
            except Exception as e:
                logger.warning(
                    f':WARN: Unable to pull groups via GraphQL. Gitlab Server version {self.client.get_api_version()}. '
                    f'Falling back to api to discover organizations. Exception: {e}'
                )
                # We may have failed partially. Err on fetching all from the same source.
                orgs = []
                for api_org in self.client.get_organizations_rest_api():
                    self.group_id_to_full_path[api_org.id] = api_org.full_path
                    orgs.append(
                        StandardizedOrganization(
                            id=str(api_org.id),
                            name=api_org.name,
                            login=str(api_org.id),
                            url=api_org.web_url,
                        )
                    )
        else:
            ingestion_type = get_ingestion_type()
            for group_id in self.config.git_organizations:
                try:
                    name, full_path, url = self.client.get_organization_name_full_path_and_url(
                        login=group_id
                    )
                    self.group_id_to_full_path[group_id] = full_path
                    orgs.append(
                        StandardizedOrganization(id=group_id, name=name, login=group_id, url=url)
                    )
                except Exception as e:
                    if ingestion_type == IngestionType.AGENT:
                        response_code = getattr(e, 'response_code', '')
                        logger.warning(
                            f'Got {type(e).__name__} {response_code} when fetching data for group {group_id}'
                        )
                    else:
                        raise

        return orgs

    def get_users(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedUser, None, None]:
        """
        Get all users in a given Git Organization

        Args:
            standardized_organization (StandardizedOrganization): A standardized Git Organization Object
            limit (int, optional): When provided, only returns this many users. Defaults to None.

        Returns:
            Generator[StandardizedUser, None, None]: A generator of StandardizedUser objects
        """
        for idx, user in enumerate(self.client.get_users(group_id=standardized_organization.login)):
            if limit and idx >= limit:
                return

            yield self._standardize_user(user)

    def get_teams(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedTeam, None, None]:
        """
        This function is to align with what the parent adapter class expects.
        GitLab does not have a concept of teams past groups, which we use as organizations.
        This will return an empty list, regardless of arguments.
        """
        teams: List[StandardizedTeam] = []
        yield from teams

    def get_repos(
        self,
        standardized_organization: StandardizedOrganization,
        limit: Optional[int] = None,
        only_private: bool = False,
    ) -> Generator[StandardizedRepository, None, None]:
        """Get a list of standardized repositories within a given organization. Currently uses
        the REST API endpoint. We have a GQL endpoint enabled in the client, but it's hard
        to get the full_path/name for each repo. If this becomes a performance bottleneck
        we should switch to the GQL endpoint

        Args:
            standardized_organization (StandardizedOrganization): A standardized organization
            limit (int, optional): When provided, only the given number of repos are returned
            only_private (bool): When True, only return private repos. Defaults to False.

        Returns:
            Generator[StandardizedRepository]: An iterable of standardized Repositories
        """
        for idx, client_repo in enumerate(
            self.client.get_repos(standardized_organization, only_private), start=1
        ):
            yield self._standardize_gitlab_project(
                client_repo,
                standardized_organization,
            )

            if limit and idx >= limit:
                return

    def get_repos_count(
        self, standardized_organization: StandardizedOrganization, only_private: bool = False
    ) -> int:
        return self.client.get_repos_count(standardized_organization.login, only_private)

    def get_commits_for_default_branch(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        """For a given repo, get all the commits that are on the Default Branch.

        Args:
            standardized_repo (StandardizedRepository): A standard Repository object
            limit (int): limit the number of commit objects we will yield
            pull_since (datetime): filter commits to be newer than this date
            pull_until (datetime): filter commits to be older than this date

        Returns:
            Generator[StandardizedCommit]: An iterable of standardized commits
        """
        default_branch_name: Optional[str] = standardized_repo.default_branch_name
        if default_branch_name:
            for j, api_commit in enumerate(
                self.client.get_commits(
                    standardized_repo,
                    since=pull_since,
                    branch_name=default_branch_name,
                ),
                start=1,
            ):
                yield self._standardize_gitlab_commit(
                    api_commit, standardized_repo, default_branch_name
                )
                if limit and j >= limit:
                    return

    def get_branches_for_repo(
        self,
        standardized_repo: StandardizedRepository,
        pull_branches: Optional[bool] = False,
        limit: Optional[int] = None,
    ) -> Generator[StandardizedBranch, None, None]:
        """Function for pulling branches for a repository. By default, pull_branches will run as False,
        so we will only process the default branch. If pull_branches is true, than we will pull all
        branches in this repository

        Args:
            standardized_repo (StandardizedRepository): A standardized repo, which hold info about the default branch.
            pull_branches (bool): A boolean flag. If True, pull all branches available on Repo.
                If false, only process the default branch. Defaults to False.
            limit (int): limit the number of branch objects we will yield

        Yields:
            StandardizedBranch: A Standardized Branch Object
        """
        branch_name: Optional[str] = None
        if not pull_branches:
            if standardized_repo.default_branch_name:
                branch_name = standardized_repo.default_branch_name
            else:
                return

        branch_count = 0

        for api_branch in self.client.get_branches_for_repo(
            standardized_repo, branch_name=branch_name
        ):
            if branch := self._standardize_gitlab_branch(
                api_branch,
                standardized_repo,
            ):
                branch_count += 1
                yield branch

            if limit and branch_count >= limit:
                return

    def get_commits_for_branches(
        self,
        standardized_repo: StandardizedRepository,
        branches: List[StandardizedBranch],
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
        limit: Optional[int] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        """For a given repo, get all the commits that are on the included branches.
        Included branches are found by crawling across the branches pulled/available
        from get_filtered_branches

        Args:
            standardized_repo (StandardizedRepository): A standard Repository object
            branches (List[StandardizedBranch]): A list of branches to pull commits from
            pull_since (datetime): A date to pull from
            pull_until (datetime): A date to pull up to
            limit (int): limit the number of commit objects we will yield

        Returns:
            List[StandardizedCommit]: A list of standardized commits
        """
        for branch in branches:
            for j, api_commit in enumerate(
                self.client.get_commits(
                    standardized_repo,
                    since=pull_since,
                    until=pull_until,
                    branch_name=str(branch.name),
                ),
                start=1,
            ):
                yield self._standardize_gitlab_commit(
                    api_commit, standardized_repo, str(branch.name)
                )

                if limit and j >= limit:
                    return

    def get_pr_metadata(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pr_pull_from_date: Optional[datetime] = None,
    ) -> Generator[StandardizedPullRequestMetadata, None, None]:
        """Get all PRs, but only included the bare necesaties

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.
            pr_pull_from_date: This is currently only used by the GithubAdapter. Probably won't be useful for this adapter

        Returns:
            List[StandardizedPullRequest]: A list of standardized PRs
        """
        raise NotImplementedError()

    def get_pr_authors(
        self, pr_list: list[StandardizedJFAPIPullRequest]
    ) -> list[StandardizedPullRequestAuthor]:
        """
        Get the authors of a list of PRs

        This is currently NOT IMPLEMENTED for GitLab instances
        """
        raise NotImplementedError()

    def get_pr_review_authors(
        self, pr_list: list[StandardizedJFAPIPullRequest]
    ) -> list[StandardizedPullRequestReviewAuthor]:
        """
        Get the review authors of a list of PRs

        This is currently NOT IMPLEMENTED for GitLab instances
        """
        raise NotImplementedError()

    def git_provider_pr_endpoint_supports_date_filtering(self) -> bool:
        """Returns a boolean on if this PR supports time window filtering.
        So far, Github DOES NOT support this (it's adapter will return False)
        but ADO does support this (it's adapter will return True)

        Returns:
            bool: A boolean on if the adapter supports time filtering when searching for PRs
        """
        return True

    def _backfill_pr_closed_at(
        self,
        standardized_repo: StandardizedRepository,
        pr_generator: Generator[Dict, None, None],
        limit: Optional[int] = None,
    ) -> Generator[Dict, None, None]:
        """
        This helper function is used after fetching prs if gql_skip_pr_closed_at is enabled via the compat config
        in the client.

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            pr_generator (Generator[Dict, None, None]): The generator to evaluate against
            limit (int, optional): When provided, limits returned prs. Passed from get_prs. Defaults to None.

        Returns:
            Generator[Dict, None, None]: A list of api prs.
        """
        evaluated_prs = {}
        # Evaluate the pr generator. Use limit to prevent pulling more than we want.
        for j, api_pr in enumerate(pr_generator, start=1):
            # IIDs are used for GitLab filtering here. They should always present and are always integers.
            evaluated_prs[int(api_pr['iid'])] = api_pr
            if limit and j >= limit:
                break

        # To fetch closed_at, use the api to fetch with batches of ids in query params
        project_id = standardized_repo.id
        project = self.client.client.projects.get(project_id)
        pr_keys = list(evaluated_prs.keys())

        # Fetch 25 pr iids at a time. This should prevent sending too many ids at a time via queryparams
        for pr_key_batch in batch_iterable(pr_keys, batch_size=25):
            for fetched_pr in project.mergerequests.list(iterator=True, iids=pr_key_batch):
                evaluated_prs[fetched_pr.iid]['closedAt'] = fetched_pr.closed_at

        for api_pr in evaluated_prs.values():
            yield api_pr

    def _backfill_pr_commits_committed_date(
        self,
        standardized_repo: StandardizedRepository,
        pr_generator: Generator[Dict, None, None],
        limit: Optional[int] = None,
    ) -> Generator[Dict, None, None]:
        """
        This helper function is used after fetching prs if gql_skip_commit_committed_date is enabled via the compat
        config in the client.

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            pr_generator (Generator[Dict, None, None]): The generator to evaluate against
            limit (int, optional): When provided, limits returned prs. Passed from get_prs. Defaults to None.

        Returns:
            Generator[Dict, None, None]: A list of api prs.
        """
        for j, api_pr in enumerate(pr_generator, start=1):
            pr_commits = self.client.get_pr_commits_rest_api(standardized_repo, api_pr)
            commit_mapping = {}
            for pr_commit in pr_commits:
                commit_mapping[pr_commit['id']] = pr_commit['committed_date']

            for gql_commit in api_pr['commits']['nodes']:
                gql_commit['committedDate'] = commit_mapping[gql_commit['sha']]

            yield api_pr

            if limit and j >= limit:
                break

    def get_prs(
        self,
        standardized_repo: StandardizedRepository,
        pull_files_for_pr: bool = False,
        hash_files_for_prs: bool = False,
        limit: Optional[int] = None,
        start_cursor: Optional[Any] = None,
        start_window: Optional[datetime] = None,
        end_window: Optional[datetime] = None,
    ) -> Generator[StandardizedPullRequest, None, None]:
        """Get the list of standardized Pull Requests for a Standardized Repository.
        This function makes use of the GraphQL interface provided by GitLab.

        Args:
            standardized_repo (StandardizedRepository): A standardized repository
            pull_files_for_pr (bool): When provided, we will pull file metadata for all PRs
            hash_files_for_prs (bool): When provided, all file metadata will be hashed for PRs
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.

        Returns:
            List[StandardizedPullRequest]: A list of standardized PRs
        """
        try:
            full_path: Optional[str] = standardized_repo.full_path
            if not full_path:
                logger.warning(
                    f':WARN: `full_path` not provided on standardized repo: {standardized_repo.name}. Skipping pulling pull requests...'
                )
                return

            pr_generator = self.client.get_prs(
                project_full_path=str(full_path),
                pull_files_for_pr=pull_files_for_pr,
                hash_files_for_pr=hash_files_for_prs,
                start_cursor=start_cursor,
                start_window=start_window,
                end_window=end_window,
            )

            if self.client.compatibility_config['gql_skip_pr_closed_at']:
                logger.info(
                    'Backfilling closed_at field for prs via api due to server GQL compatibility.'
                )
                pr_generator = self._backfill_pr_closed_at(standardized_repo, pr_generator, limit)

            if self.client.compatibility_config['gql_skip_commit_committed_date']:
                logger.info(
                    'Backfilling committed_date field for commits via api due to server GQL compatibility.'
                )
                pr_generator = self._backfill_pr_commits_committed_date(
                    standardized_repo, pr_generator, limit
                )

            for j, api_pr in enumerate(pr_generator, start=1):
                try:
                    standardized_pr = self._standardize_gitlab_pull_request_gql(
                        api_pr,
                        standardized_repo,
                    )
                    yield standardized_pr
                except (KeyError, TypeError) as e:
                    api_pr_key = None
                    if api_pr:
                        api_pr_key = api_pr.get('id')
                    logging_helper.send_to_agent_log_file(
                        f':WARN: Missing expected data for pull request key {api_pr_key}. Skipping. {e}'
                    )
                except Exception as e:
                    logging_helper.send_to_agent_log_file(
                        f'Exception encountered when attempting to standardize Pull Request. Exception: {e}',
                        level=logging.ERROR,
                    )

                if limit and j >= limit:
                    return
        except Exception as e:
            logger.warning(
                f':WARN: Unable to pull PRs for {standardized_repo.name}. Exception: {e}'
            )

    def get_users_gql(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedUser, None, None]:
        """
        NOTE: This function is not used in the current implementation due to issues with GraphQL
        authorization. It is here for future use.

        Get all users in a given Git Organization using GraphQL

        Args:
            standardized_organization (StandardizedOrganization): A standardized Git Organization Object
            limit (int, optional): When provided, the number of items returned is limited.
                Useful for the validation use case, where we want to just verify we can pull PRs.
                Defaults to None.

        Returns:
            Generator[StandardizedUser, None, None]: A generator of StandardizedUser objects
        """
        raise NotImplementedError(
            "This function is not currently supported due to GQL authorization issues."
        )

        for i, user in enumerate(
            self.client.get_users_gql(
                group_full_path=self.get_group_full_path_from_organization(
                    standardized_organization
                )
            )
        ):
            email: Optional[str] = None

            if public_email := user.get('publicEmail'):
                email = public_email
            elif email_conn_first := user.get('emails', {}).get('nodes'):
                email = email_conn_first[0].get('email')

            yield StandardizedUser(
                id=user['id'].split('gid://gitlab/User/')[1],
                name=user['name'],
                login=user['username'],
                email=email,
                url=user['webUrl'],
            )

            if limit and i >= limit:
                break



import json
import logging
import os
import shutil
import subprocess
import threading
import time
import traceback
from datetime import datetime, timedelta, timezone
from functools import wraps
from io import TextIOWrapper
from typing import NamedTuple, Optional

import psutil
import pytz
import requests

from jf_ingest import logging_helper

_DIAGNOSTICS_FILE: Optional[TextIOWrapper] = None


def _diagnostics_file_closed():
    return isinstance(_DIAGNOSTICS_FILE, TextIOWrapper) and _DIAGNOSTICS_FILE.closed


"""
The diagnostics module is used to capture diagnostics data
for agent runs. When running for Managed Ingest we do not
need to use any of this logic. For unity's sake we will
run the shared functions (so far only capture_timing),
but output nothing
"""

logger = logging.getLogger(__name__)


def _write_diagnostic(obj):
    # If there is no diagnostics file to write to, do nothing
    if not _DIAGNOSTICS_FILE:
        logger.debug(f"No diagnostics file is set")
        return

    json.dump(obj, _DIAGNOSTICS_FILE)
    _DIAGNOSTICS_FILE.write("\n")  # facilitate parsing
    _DIAGNOSTICS_FILE.flush()


def capture_agent_version():
    git_head_hash = os.getenv("SHA")
    build_timestamp = os.getenv("BUILDTIME")
    _write_diagnostic({"type": "agent_version", "sha": git_head_hash, "timestamp": build_timestamp})


# NOTE: This is the only function that is called from within the jf_ingest module!
# (as of 8/2/2023 -Gavin)
def capture_timing(*args, **kwargs):
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = pytz.utc.localize(datetime.utcnow()).isoformat()
            if "func_name_override" in kwargs:
                func_name = kwargs.pop("func_name_override")
            else:
                func_name = func.__name__
            ret = func(*args, **kwargs)
            end_time = pytz.utc.localize(datetime.utcnow()).isoformat()
            diag_obj = {
                "type": "func_timing",
                "name": func_name,
                "start": start_time,
                "end": end_time,
            }
            if ret is not None:
                try:
                    diag_obj["num_items"] = len(ret)
                except TypeError:
                    if type(ret) is int:
                        diag_obj["num_items"] = ret
            _write_diagnostic(diag_obj)
            return ret

        return wrapper

    return actual_decorator


def capture_run_args(run_mode: str, config_file: str, outdir: str, prev_output_dir: str):
    _write_diagnostic(
        {
            "type": "run_args",
            "run_mode": run_mode,
            "config_file": config_file,
            "outdir": outdir,
            "prev_output_dir": prev_output_dir,
        }
    )


def capture_outdir_size(outdir: str):
    _write_diagnostic(
        {
            "type": "outdir_size",
            "size_kb": int(
                subprocess.check_output(["du", "-sk", outdir]).split()[0].decode("utf-8")
            ),
        }
    )


class HeartBeatStart(NamedTuple):
    timestamp_key: str
    time: str  # This is an ISO timestamp
    memory_allocated_in_gb: float
    run_git: bool
    run_jira: bool
    will_send: bool


class SysReading(NamedTuple):
    """This is used as the diagnostic 'heartbeat'"""

    timestamp_key: str
    time: str  # This is an ISO timestamp
    cpu_pct: float
    memory_used_gb: float
    memory_pct: float
    disk_used_gb: float
    disk_pct: float


class HeartBeatEnd(NamedTuple):
    timestamp_key: str
    time: str  # This is an ISO timestamp
    git_success: bool  # NOTE / TODO: You can have multiple git instances, but this boolean reflects all or nothing success
    jira_success: bool


def post_to_agent_monitor(url: str, jellyfish_api_token: str, data: NamedTuple):
    try:
        headers = {'X-JF-API-Token': jellyfish_api_token}

        r = requests.post(
            url,
            headers=headers,
            json=data._asdict(),
        )

        r.raise_for_status()
    except Exception as e:
        logging_helper.send_to_agent_log_file(
            f'Failed to send initial Diagnostic heart beat to Jellyfish. Error: {e}'
        )
        logging_helper.send_to_agent_log_file(traceback.format_exc(), level=logging.ERROR)


def send_diagnostic_start_reading(
    jellyfish_webhook_base: str,
    jellyfish_api_token: str,
    timestamp_key: str,
    will_run_git: bool,
    will_run_jira: bool,
    will_send: bool,
):
    heart_beat_start = HeartBeatStart(
        memory_allocated_in_gb=psutil.virtual_memory().total / (1024**3),
        timestamp_key=timestamp_key,
        time=datetime.now(timezone.utc).isoformat(),
        run_git=will_run_git,
        run_jira=will_run_jira,
        will_send=will_send,
    )
    post_to_agent_monitor(
        url=f'{jellyfish_webhook_base}/agent-monitor/start',
        jellyfish_api_token=jellyfish_api_token,
        data=heart_beat_start,
    )


def send_diagnostic_end_reading(
    jellyfish_webhook_base: str,
    jellyfish_api_token: str,
    timestamp_key: str,
    git_success: bool,
    jira_success: bool,
):
    heart_beat_end = HeartBeatEnd(
        timestamp_key=timestamp_key,
        time=datetime.now(timezone.utc).isoformat(),
        git_success=git_success,
        jira_success=jira_success,
    )
    post_to_agent_monitor(
        url=f'{jellyfish_webhook_base}/agent-monitor/stop',
        jellyfish_api_token=jellyfish_api_token,
        data=heart_beat_end,
    )


def continually_gather_system_diagnostics(
    kill_event: threading.Event,
    outdir: str,
    jellyfish_webhook_base: str,
    jellyfish_api_token: str,
    send_to_jf_endpoint: bool = True,
    cycle_period_in_seconds: int = 60,
):
    def _flush_cached_readings(cached_readings: list[SysReading]):
        if not cached_readings:
            return
        _write_diagnostic(
            {
                "type": "sys_resources_60s",
                "start": cached_readings[0].time,
                "end": cached_readings[-1].time,
                "cpu_pct": ["%.2f" % r.cpu_pct for r in cached_readings],
                "mem_used_gb": ["%.2f" % r.memory_used_gb for r in cached_readings],
                "mem_pct": ["%.2f" % r.memory_pct for r in cached_readings],
                "disk_used_gb": ["%.2f" % r.disk_used_gb for r in cached_readings],
                "disk_pct": ["%.2f" % r.disk_pct for r in cached_readings],
            }
        )

    readings = threading.local()
    readings.cached_readings = []
    readings.last_reading_time = None
    readings.last_flush_time = datetime.now(timezone.utc)

    while True:
        if kill_event.is_set():
            _flush_cached_readings(readings.cached_readings)
            readings.cached_readings = []
            return
        else:
            now = datetime.now(timezone.utc)
            if not readings.last_reading_time or (now - readings.last_reading_time) > timedelta(
                seconds=cycle_period_in_seconds
            ):
                # First, gather diagnostics data
                cpu = psutil.cpu_percent()
                memory = psutil.virtual_memory()
                disk = shutil.disk_usage(outdir)
                _, timestamp_key = os.path.split(outdir)
                sys_reading = SysReading(
                    timestamp_key,
                    now.isoformat(),
                    cpu / 100,
                    (memory.total - memory.available) / (1024**3),
                    (memory.total - memory.available) / memory.total,
                    disk.used / (1024**3),
                    disk.used / disk.total,
                )
                # Next submit diagnostics data to the local cache to eventually
                # flush to the diagnostics.json
                readings.cached_readings.append(sys_reading)
                readings.last_reading_time = now

                # Finally, send this snapshot of diagnostics reading to the jellyfish
                # server as a 'heartbeat' diagnostics update if enabled
                if send_to_jf_endpoint:
                    post_to_agent_monitor(
                        url=f'{jellyfish_webhook_base}/agent-monitor/heartbeat',
                        jellyfish_api_token=jellyfish_api_token,
                        data=sys_reading,
                    )

            if now - readings.last_flush_time > timedelta(seconds=300):
                _flush_cached_readings(readings.cached_readings)
                readings.cached_readings = []
                readings.last_flush_time = now

            # Keep the sleep short so that the thread's responsive to the kill_event
            time.sleep(1)


def open_file(outdir: str):
    global _DIAGNOSTICS_FILE
    try:
        if _DIAGNOSTICS_FILE is None:
            _DIAGNOSTICS_FILE = open(os.path.join(outdir, "diagnostics.json"), "a")
        elif _diagnostics_file_closed():
            _DIAGNOSTICS_FILE = open(os.path.join(outdir, "diagnostics.json"), "a")
        else:
            logger.debug(f"Diagnostics file is already open")
    except Exception as e:
        # These messages are left at the DEBUG level for a better UX in agent
        logger.debug(f"Error opening Diagnostics file. Diagnostics will not be written. Error: {e}")


def close_file():
    try:
        global _DIAGNOSTICS_FILE
        if _DIAGNOSTICS_FILE is None:
            logger.debug("Diagnostics file is already closed")
        elif _diagnostics_file_closed():
            logger.debug("Diagnostics file is already closed")
        else:
            logger.debug("Closing diagnostics file")
            _DIAGNOSTICS_FILE.close()
        _DIAGNOSTICS_FILE = None
    except Exception as e:
        # These messages are left at the DEBUG level for a better UX in agent
        logger.debug(f"Error closing Diagnostics file. Diagnostics will not be written. Error: {e}")



import datetime
import json
import string
from contextlib import contextmanager

import pytest
import requests_mock

from jf_ingest.jf_jira.downloaders import (
    _get_all_users_for_gdpr_active_instance,
    _scrub_jira_users,
    _search_users_by_letter,
    _should_augment_email,
    augment_jira_users_with_email,
    download_users,
    get_searchable_jira_letters,
    search_users,
)
from jf_ingest.jf_jira.exceptions import NoJiraUsersFoundException
from jf_ingest.utils import batch_iterable
from tests.jf_jira.utils import _register_jira_uri, get_jira_mock_connection

DEFAULT_PAGE_SIZE = 1000


def _generate_spoofed_user_data(user_count: int, char: str = "default", gdpr_active: bool = True):
    return [
        {
            f"{'accountId' if gdpr_active else 'key'}": f"{char}_{x}",
            "emailAddress": f"{char}_email_{x}@emailAddress.com",
        }
        for x in range(0, user_count)
    ]


@contextmanager
def _spoof_gdpr_server(page_size: int, total_users: int, server_page_size: int = DEFAULT_PAGE_SIZE):
    # Test with incremental paging
    with requests_mock.Mocker() as m:
        # SET UP MOCKED USER DATA
        lower_index = 0
        user_chunk = []
        user_data = _generate_spoofed_user_data(total_users)
        for user_chunk in batch_iterable(user_data, batch_size=page_size):
            _register_jira_uri(
                m,
                f"users/search?startAt={lower_index}&maxResults={page_size}",
                json.dumps(user_chunk),
            )
            lower_index += len(user_chunk)
        # Get last page of data
        _register_jira_uri(
            m,
            f"users/search?startAt={total_users}&maxResults={page_size}",
            json.dumps([]),
        )
        yield user_data


@contextmanager
def _mock_search_by_user_api_that_can_page(
    test_data: dict, page_size: int, returned_server_page_size: int = DEFAULT_PAGE_SIZE
):
    with requests_mock.Mocker() as m:
        for letter in get_searchable_jira_letters():
            # Load up empty values for every letter as a baseline
            _register_jira_uri(
                m,
                f"user/search?startAt=0&maxResults={page_size}&includeActive=True&includeInactive=True&username={letter}",
                "[]",
            )

        for char, data in test_data.items():
            # Overwrite the letter a to return results
            start_at = 0
            for user_batch in batch_iterable(data, returned_server_page_size):
                _register_jira_uri(
                    m,
                    f"user/search?startAt={start_at}&maxResults={page_size}&includeActive=True&includeInactive=True&username={char}",
                    json.dumps(user_batch),
                )
                start_at += len(user_batch)
            _register_jira_uri(
                m,
                f"user/search?startAt={start_at}&maxResults={page_size}&includeActive=True&includeInactive=True&username={char}",
                json.dumps([]),
            )

            for letter in get_searchable_jira_letters():
                # For the use case where we have exactly a pages worth of data, we
                # will iterate on the second character. To make this wrapper cover all
                # cases, set a baseline that sets every endpoint for a recursive depth of one (aa, ab, ac, ...)
                _register_jira_uri(
                    m,
                    f"user/search?startAt=0&maxResults={page_size}&includeActive=True&includeInactive=True&username={char}{letter}",
                    "[]",
                )
        yield


def test_search_users_data_less_than_page_size():
    test_data = {
        "a": _generate_spoofed_user_data(
            char="a", user_count=DEFAULT_PAGE_SIZE // 2, gdpr_active=False
        )
    }
    with _mock_search_by_user_api_that_can_page(test_data=test_data, page_size=DEFAULT_PAGE_SIZE):
        users = _search_users_by_letter(
            get_jira_mock_connection(),
            False,
            search_users_by_letter_email_domain=None,
            page_size=DEFAULT_PAGE_SIZE,
        )
        assert type(users) == list
        assert users == test_data["a"]


def test_search_users_data_exact_page_size():
    test_data = {
        "a": _generate_spoofed_user_data(char="a", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False)
    }
    with _mock_search_by_user_api_that_can_page(test_data=test_data, page_size=DEFAULT_PAGE_SIZE):
        users = _search_users_by_letter(
            get_jira_mock_connection(),
            False,
            search_users_by_letter_email_domain=None,
            page_size=DEFAULT_PAGE_SIZE,
        )
        assert type(users) == list
        assert users == test_data["a"]


def test_search_users_data_larger_than_page_size():
    test_data = {
        "a": _generate_spoofed_user_data(
            char="a",
            user_count=DEFAULT_PAGE_SIZE + (DEFAULT_PAGE_SIZE * 2 + 523),
            gdpr_active=False,
        )
    }
    with _mock_search_by_user_api_that_can_page(test_data=test_data, page_size=DEFAULT_PAGE_SIZE):
        users = _search_users_by_letter(
            get_jira_mock_connection(),
            False,
            search_users_by_letter_email_domain=None,
            page_size=DEFAULT_PAGE_SIZE,
        )
        assert type(users) == list
        assert len(users) == len(test_data["a"])
        assert users == test_data["a"]


def test_search_users_data_using_numeric():
    test_data = {
        "7": _generate_spoofed_user_data(
            char="7",
            user_count=DEFAULT_PAGE_SIZE - DEFAULT_PAGE_SIZE // 3,
            gdpr_active=False,
        )
    }
    with _mock_search_by_user_api_that_can_page(test_data=test_data, page_size=DEFAULT_PAGE_SIZE):
        users = _search_users_by_letter(
            get_jira_mock_connection(),
            False,
            search_users_by_letter_email_domain=None,
            page_size=DEFAULT_PAGE_SIZE,
        )
        assert type(users) == list
        assert len(users) == len(test_data["7"])
        assert users == test_data["7"]


def test_search_users_data_all_test_cases():
    test_data = {
        "a": _generate_spoofed_user_data(
            char="a", user_count=DEFAULT_PAGE_SIZE // 2, gdpr_active=False
        ),
        "f": _generate_spoofed_user_data(char="f", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False),
        "g": _generate_spoofed_user_data(
            char="g",
            user_count=DEFAULT_PAGE_SIZE + (DEFAULT_PAGE_SIZE * 2 + 523),
            gdpr_active=False,
        ),
        "7": _generate_spoofed_user_data(
            char="7", user_count=DEFAULT_PAGE_SIZE + 685, gdpr_active=False
        ),
    }

    with _mock_search_by_user_api_that_can_page(test_data=test_data, page_size=DEFAULT_PAGE_SIZE):
        users = _search_users_by_letter(
            get_jira_mock_connection(),
            False,
            search_users_by_letter_email_domain=None,
            page_size=DEFAULT_PAGE_SIZE,
        )

        assert type(users) == list

        flattened_test_data = []
        for data in test_data.values():
            flattened_test_data.extend(data)
        assert len(users) == len(flattened_test_data)
        assert users == flattened_test_data


def test_search_users_data_all_test_cases_and_special_server_batching():
    test_data = {
        "a": _generate_spoofed_user_data(
            char="a", user_count=DEFAULT_PAGE_SIZE // 2, gdpr_active=False
        ),
        "f": _generate_spoofed_user_data(char="f", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False),
        "g": _generate_spoofed_user_data(
            char="g",
            user_count=DEFAULT_PAGE_SIZE + (DEFAULT_PAGE_SIZE * 2 + 523),
            gdpr_active=False,
        ),
        "7": _generate_spoofed_user_data(
            char="7", user_count=DEFAULT_PAGE_SIZE + 685, gdpr_active=False
        ),
    }

    with _mock_search_by_user_api_that_can_page(
        test_data=test_data, page_size=DEFAULT_PAGE_SIZE, returned_server_page_size=100
    ):
        users = _search_users_by_letter(
            get_jira_mock_connection(),
            False,
            search_users_by_letter_email_domain=None,
            page_size=DEFAULT_PAGE_SIZE,
        )

        assert type(users) == list

        flattened_test_data = []
        for data in test_data.values():
            flattened_test_data.extend(data)
        assert len(users) == len(flattened_test_data)
        assert users == flattened_test_data


def test_search_users_data_recursive_case():
    test_data = {
        # exactly page size and nothing more
        "a": _generate_spoofed_user_data(char="a", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False),
        # Page size with some extra
        "b": _generate_spoofed_user_data(char="b", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False),
        "ba": _generate_spoofed_user_data(
            char="ba", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False
        ),
        "bar": _generate_spoofed_user_data(
            char="bar", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False
        ),
        "barb": _generate_spoofed_user_data(
            char="barb", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False
        ),
        # Test some numbers
        "4": _generate_spoofed_user_data(char="4", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False),
        "42": _generate_spoofed_user_data(
            char="42", user_count=DEFAULT_PAGE_SIZE, gdpr_active=False
        ),
        "426": _generate_spoofed_user_data(
            char="426",
            user_count=DEFAULT_PAGE_SIZE // 2,
            gdpr_active=False,
        ),
        "429": _generate_spoofed_user_data(
            char="4267",
            user_count=DEFAULT_PAGE_SIZE // 3,
            gdpr_active=False,
        ),
        "8": _generate_spoofed_user_data(
            char="8", user_count=DEFAULT_PAGE_SIZE // 4, gdpr_active=False
        ),
    }

    with _mock_search_by_user_api_that_can_page(test_data=test_data, page_size=DEFAULT_PAGE_SIZE):
        users = _search_users_by_letter(
            get_jira_mock_connection(),
            False,
            search_users_by_letter_email_domain=None,
            page_size=DEFAULT_PAGE_SIZE,
        )

        assert type(users) == list

        flattened_test_data = []
        for data in test_data.values():
            flattened_test_data.extend(data)
        assert len(users) == len(flattened_test_data)
        assert users == flattened_test_data


def test_get_all_users_for_gdpr_active_instance():
    page_size = DEFAULT_PAGE_SIZE
    total_users = 1668
    with _spoof_gdpr_server(page_size=page_size, total_users=total_users) as yielded_user_data:
        # Test 'base level' paging
        user_data = _get_all_users_for_gdpr_active_instance(
            get_jira_mock_connection(), page_size=page_size
        )
        assert len(user_data) == total_users

        # test wrapper function
        users_full_test = search_users(
            jira_connection=get_jira_mock_connection(),
            gdpr_active=True,
            page_size=page_size,
        )
        assert len(users_full_test) == total_users

        assert users_full_test == yielded_user_data

        users_check_augmentation = download_users(
            jira_basic_connection=get_jira_mock_connection(),
            jira_atlas_connect_connection=None,
            gdpr_active=True,
        )

        for user in users_check_augmentation:
            assert "email_pulled" in user


def test_get_all_users_for_gdpr_active_instance_with_non_max_results_paging():
    page_size = DEFAULT_PAGE_SIZE
    total_users = 1668
    with _spoof_gdpr_server(
        page_size=page_size, total_users=total_users, server_page_size=23
    ) as yielded_user_data:
        # Test 'base level' paging
        user_data = _get_all_users_for_gdpr_active_instance(
            get_jira_mock_connection(), page_size=page_size
        )
        assert len(user_data) == total_users

        # test wrapper function
        users_full_test = search_users(
            jira_connection=get_jira_mock_connection(),
            gdpr_active=True,
            page_size=page_size,
        )
        assert len(users_full_test) == total_users

        assert users_full_test == yielded_user_data

        users_check_augmentation = download_users(
            jira_basic_connection=get_jira_mock_connection(),
            jira_atlas_connect_connection=None,
            gdpr_active=True,
        )

        for user in users_check_augmentation:
            assert "email_pulled" in user


def test_get_searchable_jira_letters():
    # Due to some Jira API weirdness, we NEVER want to hit the API endpoint
    # with non letter or non digit characters.
    for letter in get_searchable_jira_letters():
        assert letter not in string.punctuation
        assert letter not in string.whitespace


def test_download_users():
    page_size = DEFAULT_PAGE_SIZE
    total_users = 1668
    with _spoof_gdpr_server(page_size=page_size, total_users=total_users) as yielded_user_data:
        # Test 'base level' paging
        user_data = download_users(
            jira_basic_connection=get_jira_mock_connection(),
            jira_atlas_connect_connection=None,
            gdpr_active=True,
            search_users_by_letter_email_domain=None,
            required_email_domains=[],
            is_email_required=False,
        )

        assert len(user_data) == total_users

        for user in user_data:
            assert "email_pulled" in user


def test_download_users_filter_by_required_email_domains():
    page_size = DEFAULT_PAGE_SIZE
    total_users = 1668
    with _spoof_gdpr_server(page_size=page_size, total_users=total_users):
        # Test 'base level' paging
        with pytest.raises(NoJiraUsersFoundException):
            user_data = download_users(
                jira_basic_connection=get_jira_mock_connection(),
                jira_atlas_connect_connection=None,
                gdpr_active=True,
                search_users_by_letter_email_domain=None,
                required_email_domains=["donotinclude.com"],
                is_email_required=False,
            )

        user_data = download_users(
            jira_basic_connection=get_jira_mock_connection(),
            jira_atlas_connect_connection=None,
            gdpr_active=True,
            search_users_by_letter_email_domain=None,
            required_email_domains=["emailAddress.com"],
            is_email_required=False,
        )
        assert len(user_data) == total_users


def test_scrub_users():
    mocked_users = [
        {"emailAddress": "test@domainOne.com"},
        {"emailAddress": "test@domainTwo.com"},
        {"emailAddress": "test@domainThree.com"},
    ]

    scrubbed_users = _scrub_jira_users(
        mocked_users,
        required_email_domains=["NoDomainFound.com"],
        is_email_required=False,
    )
    assert len(scrubbed_users) == 0

    scrubbed_users = _scrub_jira_users(
        mocked_users, required_email_domains=["domainOne.com"], is_email_required=False
    )
    assert len(scrubbed_users) == 1

    scrubbed_users = _scrub_jira_users(
        mocked_users,
        required_email_domains=["domainOne.com", "domainTwo.com"],
        is_email_required=False,
    )
    assert len(scrubbed_users) == 2

    scrubbed_users = _scrub_jira_users(
        mocked_users,
        required_email_domains=["domainOne.com", "domainTwo.com", "domainThree.com"],
        is_email_required=False,
    )
    assert len(scrubbed_users) == 3

    scrubbed_users = _scrub_jira_users(
        mocked_users, required_email_domains=[], is_email_required=False
    )
    assert len(scrubbed_users) == 3


def test_augment_with_email():
    # Test with incremental paging
    with requests_mock.Mocker() as m:
        jira_users = _generate_spoofed_user_data(55)
        augmented_users = set()
        for user in jira_users:
            account_id = user["accountId"]
            if _should_augment_email(user):
                _register_jira_uri(
                    m,
                    f"user/email?accountId={account_id}",
                    f'{{"accountId": "{account_id}", "email": "dummy@email.com"}}',
                )
                augmented_users.add(account_id)

        augment_jira_users_with_email(get_jira_mock_connection(), jira_users)
        for user in jira_users:
            if user["accountId"] in augmented_users:
                assert user["emailAddress"] == "dummy@email.com"
                assert "email_pulled" in user
                assert isinstance(user["email_pulled"], datetime.datetime)



import json
import logging
from dataclasses import asdict
from datetime import datetime
from enum import Enum
from typing import Generator, Optional

from jira import JIRA

from jf_ingest import diagnostics, logging_helper
from jf_ingest.config import (
    IngestionConfig,
    IngestionType,
    IssueDownloadingResult,
    JiraAuthMethod,
    JiraDownloadConfig,
    UserMetadata,
)
from jf_ingest.constants import Constants
from jf_ingest.events.models import IngestType, JiraIngestEvent
from jf_ingest.file_operations import IngestIOHelper, SubDirectory
from jf_ingest.jf_jira.auth import get_jira_connection
from jf_ingest.jf_jira.downloaders import (
    download_boards_and_sprints,
    download_fields,
    download_global_components,
    download_issuelinktypes,
    download_issuetypes,
    download_priorities,
    download_projects_and_versions_and_components,
    download_resolutions,
    download_statuses,
    download_users,
    download_users_by_urls,
    download_worklogs,
    get_issue_ids_for_rekeys_and_deletes,
    get_jira_search_batch_size,
    is_jql_enhanced_search_available,
    pull_all_jira_issues_by_date,
    pull_jira_issues_by_jira_ids,
)
from jf_ingest.jf_jira.exceptions import AtlassianConnectException
from jf_ingest.jf_jira.utils import (
    JiraFieldIdentifier,
    expand_and_normalize_jira_fields,
    get_unique_users_from_issue,
    get_user_key,
)
from jf_ingest.telemetry import add_telemetry_fields, jelly_trace, record_span
from jf_ingest.utils import (
    batch_iterable_by_bytes_size,
    get_jellyfish_company_slug,
    get_object_bytes_size,
    init_jf_ingest_run,
)

logger = logging.getLogger(__name__)


def _generate_jira_event(event_name: str) -> JiraIngestEvent:
    return JiraIngestEvent(
        company_slug=get_jellyfish_company_slug(),
        event_name=event_name,
        ingest_type=IngestType.JIRA,
    )


class JiraObject(Enum):
    JiraFields = "jira_fields"
    JiraProjectsAndVersions = "jira_projects_and_versions"
    JiraGlobalComponents = "jira_global_components"
    JiraUsers = "jira_users"
    JiraResolutions = "jira_resolutions"
    JiraIssueTypes = "jira_issuetypes"
    JiraLinkTypes = "jira_linktypes"
    JiraPriorities = "jira_priorities"
    JiraBoards = "jira_boards"
    JiraSprints = "jira_sprints"
    JiraBoardSprintLinks = "jira_board_sprint_links"
    JiraIssues = "jira_issues"
    JiraIssuesIdsDownloaded = "jira_issue_ids_downloaded"
    JiraIssuesIdsSkipped = "jira_issue_ids_skipped"
    JiraIssuesIdsDeleted = "jira_issue_ids_deleted"
    JiraWorklogs = "jira_worklogs"
    JiraStatuses = "jira_statuses"
    JiraDownloadMetadata = "jira_download_metadata"


@jelly_trace()
def load_issues_in_batches(
    issues_to_download: Generator[dict, None, None],
    ingest_io_helper: IngestIOHelper,
    ingest_config: IngestionConfig,
    jira_config: JiraDownloadConfig,
    batch_number_start: int = 0,
) -> IssueDownloadingResult:
    """given a generator object for issues to download; save them to disk in batches according to maximum size

    Args:
        issues_to_download: Generator object for issues to download
        ingest_io_helper: IngestIOHelper object for doing file operations
        ingest_config (IngestionConfig): A dataclass that holds several different configuration args for this task
        batch_number_start: starting index for the batch number (jira_issuesN.json). Used if you call this function multiple times

    Return:
        IssueDownloadingResult: A dataclass that holds the results of the download;
            - the ids for jira issues retrieved,
            - jira_ids for parents of those retrieved issues,
            - and the total number of batches it took
    """
    total_issue_batches = 0
    all_downloaded_issue_ids = set()
    collected_parent_ids = set()
    issues_ids_too_large_to_download = set()
    unique_users_across_all_issues = set()

    def _filter_out_large_issues(
        issues_generator: Generator[dict, None, None]
    ) -> Generator[dict, None, None]:
        """When processing issues we should not process any issue that's larger than 255 MB in it's string form.
        This is a hard limit we have on issue insertion logic.
        """
        for issue in issues_generator:
            if get_object_bytes_size(json.dumps(issue)) >= Constants.JIRA_ISSUE_SIZE_LIMIT:
                issues_ids_too_large_to_download.add(issue['id'])
                continue
            yield issue

    def _extract_unique_users_within_issues(
        issues_generator: Generator[dict, None, None]
    ) -> Generator[dict, None, None]:
        for issue in issues_generator:
            unique_users_across_all_issues.update(
                get_unique_users_from_issue(issue, gdpr_active=jira_config.gdpr_active)
            )
            yield issue

    issues_to_download = _filter_out_large_issues(issues_generator=issues_to_download)
    if jira_config.feature_flags.get(Constants.CHECK_ISSUES_FOR_EXTRA_USERS_AND_PULL):
        issues_to_download = _extract_unique_users_within_issues(
            issues_generator=issues_to_download
        )
    for batch_number, batch_issues in enumerate(
        batch_iterable_by_bytes_size(
            issues_to_download, batch_byte_size=Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE
        ),
        start=batch_number_start,
    ):
        logging_helper.send_to_agent_log_file(
            f'Saving {len(batch_issues)} issues as batch number {batch_number}', level=logging.DEBUG
        )
        all_downloaded_issue_ids.update(set(issue['id'] for issue in batch_issues))
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssues.value,
            subdirectory=SubDirectory.JIRA,
            json_data=batch_issues,
            batch_number=batch_number,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )
        total_issue_batches += 1

        # we need to download parents of issues we just fetched
        collected_parent_ids.update(
            set(
                issue['fields']['parent']['id']
                for issue in batch_issues
                if 'parent' in issue['fields'].keys()
            )
        )

    logger.info(
        f"Successfully saved {len(all_downloaded_issue_ids)} Jira Issues in "
        f"{total_issue_batches} separate batches, with each batch limited to "
        f"{round(Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE / Constants.MB_SIZE_IN_BYTES)}MB per batch"
    )

    return IssueDownloadingResult(
        discovered_parent_ids=collected_parent_ids,
        issue_ids_too_large_to_upload=issues_ids_too_large_to_download,
        downloaded_ids=all_downloaded_issue_ids,
        total_batches=total_issue_batches,
        users_found=unique_users_across_all_issues,
    )


@jelly_trace()
def pull_issues_wrapper(
    jira_connect_or_fallback_connection: JIRA,
    jira_config: JiraDownloadConfig,
    projects_and_versions: list[dict],
    include_fields: list[JiraFieldIdentifier],
    exclude_fields: list[JiraFieldIdentifier],
    jira_issues_batch_size: int,
    ingest_io_helper: IngestIOHelper,
    ingest_config: IngestionConfig,
    use_jql_enhanced_search: bool = False,
) -> tuple[set[str], set[str], set[UserMetadata]]:
    """Pulls issues using the "new" sync path, based on our ability to pull 10,000 jira issue IDs at one time
        as long as you only ask for `id` and `key` in the returned fields. Because we get no more fields we detect
        deletes as the absence of an issue on remote which we have local, and key changes as id match key mismatch
    Args:
        jira_connect_or_fallback_connection: either the basic connection or the special atlassian connect keys
        jira_config: jira config for jf_ingest from JF
        projects_and_versions: list of projects and versions to pull issues from
        jira_issues_batch_size: batch size to use for actually downloading issues (much lower than the 10k for IDs)
        ingest_io_helper: IngestIOHelper object for doing file operations
        ingest_config: full jf_ingest config
        use_jql_enhanced_search: whether to use JQL Enhanced Search API (/search/jql) or legacy API (/search)

    Returns:
        tuple[set[str], set[str], set[UserMetadata]]: returns a tuple of three sets:
            The first set is all the Jira Issue IDs that we downloaded
            The second set is all the Jira IDs that have been deleted
            The third set is all the unique user objects across every issue
    """

    logging_helper.send_to_agent_log_file("Running 10k batched sync path")
    logger.info(
        f"Using {jira_issues_batch_size} as batch size for issue downloads. Full Redownload is {jira_config.full_redownload}"
    )
    issue_ids_not_uploaded: set[str] = set()
    ###########################################################################
    #
    # STEP ONE: Crawl across all projects to pull all Issue ID and Keys
    # to detect potential deletions, rekeys, or creations
    #
    ###########################################################################
    with record_span('pull_jira_issues_crawl'):
        issue_list_diff = get_issue_ids_for_rekeys_and_deletes(
            jira_connection=jira_connect_or_fallback_connection,
            jellyfish_issue_metadata=jira_config.jellyfish_issue_metadata,
            project_key_to_id={
                proj["key"]: proj["id"]
                for proj in projects_and_versions
                if proj["key"] not in jira_config.exclude_projects
            },
            pull_from=jira_config.pull_from,
            jql_filter=jira_config.issue_jql_filter,
            use_jql_enhanced_search=use_jql_enhanced_search,
        )
        ids_to_delete = issue_list_diff.ids_to_delete
        ids_to_download = issue_list_diff.ids_to_download

    ###########################################################################
    #
    # STEP TWO [Optional]: Pull issues by their updated date
    #
    ###########################################################################
    issues_by_updated_date_results: IssueDownloadingResult = IssueDownloadingResult(
        downloaded_ids=set(),
        issue_ids_too_large_to_upload=set(),
        discovered_parent_ids=set(),
        total_batches=0,
        users_found=set(),
    )
    if jira_config.pull_issues_by_date:
        project_id_to_pull_from: dict[str, datetime] = {
            str(proj["id"]): (
                jira_config.project_id_to_pull_from.get(str(proj["id"]), jira_config.pull_from)
                if not jira_config.full_redownload and jira_config.project_id_to_pull_from
                else jira_config.pull_from
            )
            for proj in projects_and_versions
            if proj["key"] not in jira_config.exclude_projects
        }
        logger.info(
            f'Attempting to pull all the full Jira Issues that have been updated since our last Jira Download across {len(project_id_to_pull_from)} projects.'
        )
        issues_by_updated_date_generator = pull_all_jira_issues_by_date(
            jira_connection=jira_connect_or_fallback_connection,
            project_key_to_pull_from=project_id_to_pull_from,
            num_parallel_threads=jira_config.issue_download_concurrent_threads,
            batch_size=jira_issues_batch_size,
            jql_project_batch_size=50,
            jql_filter=jira_config.issue_jql_filter,
            expand_fields=["changelog"],
            include_fields=include_fields,
            exclude_fields=exclude_fields,
            use_jql_enhanced_search=use_jql_enhanced_search,
        )
        with record_span('pull_jira_issues_by_date'):
            issues_by_updated_date_results = load_issues_in_batches(
                issues_by_updated_date_generator,
                ingest_io_helper,
                ingest_config,
                ingest_config.jira_config,
            )
            issue_ids_not_uploaded.update(
                issues_by_updated_date_results.issue_ids_too_large_to_upload
            )
            logger.info(
                f'We have downloaded {len(issues_by_updated_date_results.downloaded_ids)} issues that have been updated or created since our last pull across {len(project_id_to_pull_from)} projects.'
            )

    file_batch_cursor = issues_by_updated_date_results.total_batches
    ###########################################################################
    #
    # Step 3: Download the full Jira Issues for potentially rekeys and/or missing issues.
    # Also redownload issues that have been marked for redownload in our system
    #
    ###########################################################################
    # Extend this list with issues that need to be redownloaded for Jellyfish reasons
    ids_to_download.update(jira_config.jellyfish_issue_ids_for_redownload)
    # Next, exclude any IDs that we have just done a download for, via the updated date
    ids_to_download = ids_to_download.difference(issues_by_updated_date_results.downloaded_ids)
    logging_helper.send_to_agent_log_file(
        f'We need to download an additional {len(ids_to_download)} issues because they have been rekeyed or marked for redownload by Jellyfish'
    )

    # Finally, we have a "shopping list" of IDs to download
    issues_by_ids_generator = pull_jira_issues_by_jira_ids(
        jira_connection=jira_connect_or_fallback_connection,
        jira_ids=ids_to_download,
        num_parallel_threads=jira_config.issue_download_concurrent_threads,
        batch_size=jira_issues_batch_size,
        expand_fields=["changelog"],
        include_fields=include_fields,
        exclude_fields=exclude_fields,
        use_jql_enhanced_search=use_jql_enhanced_search,
    )
    with record_span('pull_jira_issues_by_ids'):
        issues_by_ids_results: IssueDownloadingResult = load_issues_in_batches(
            issues_by_ids_generator,
            ingest_io_helper,
            ingest_config,
            ingest_config.jira_config,
            batch_number_start=file_batch_cursor,
        )
        file_batch_cursor += issues_by_ids_results.total_batches
        issue_ids_not_uploaded.update(issues_by_ids_results.issue_ids_too_large_to_upload)
    # Now, combine all issue IDs that we've downloaded
    all_downloaded_ids: set[str] = set()
    all_downloaded_ids.update(issues_by_updated_date_results.downloaded_ids)
    all_downloaded_ids.update(issues_by_ids_results.downloaded_ids)

    # Then, combine the parents of the issues we downloaded by update date, and issues we downloaded by ID
    all_discovered_parent_ids: set[str] = set()
    all_discovered_parent_ids.update(issues_by_updated_date_results.discovered_parent_ids)
    all_discovered_parent_ids.update(issues_by_ids_results.discovered_parent_ids)

    # Now, find unique parents that we have NOT downloaded and that we need to
    existing_ids = set([jimd.id for jimd in jira_config.jellyfish_issue_metadata])
    all_known_ids = existing_ids.union(all_downloaded_ids)
    to_be_downloaded_parent_ids = all_discovered_parent_ids.difference(all_known_ids)

    # Track all unique users across all issues we're downloading
    users_found_across_all_issues: set[UserMetadata] = set()
    users_found_across_all_issues.update(issues_by_updated_date_results.users_found)
    users_found_across_all_issues.update(issues_by_ids_results.users_found)

    ###########################################################################
    #
    # STEP FOUR: fetch parents
    # NOTE: For the general Jellyfish use case, we only need to go "one level deep" on parents,
    # i.e. pull the set of missing parents but NOT the parents of those parents (the "grandparents").
    # There is an optional feature, however, that will allow us to pull the parents of parents of parents
    # until we've pulled all parents. This is NOT the general use case, and will likely add a lot of computation
    # time to this function, so we advise you only use it in very specific use cases
    #
    ###########################################################################
    with record_span('pull_jira_issues_parents'):
        depth_level = 1
        while to_be_downloaded_parent_ids:
            logging_helper.send_to_agent_log_file(
                f'Attempting to pull more parents ({len(to_be_downloaded_parent_ids)} detected parents to pull). Some of them are {sorted(list(to_be_downloaded_parent_ids))[:30]} Depth level: {depth_level}. Any parents before {jira_config.pull_from} will not be downloaded'
            )
            # Finally, we have a "shopping list" of IDs to download
            issues_by_ids_generator = pull_jira_issues_by_jira_ids(
                jira_connection=jira_connect_or_fallback_connection,
                jira_ids=to_be_downloaded_parent_ids,
                num_parallel_threads=jira_config.issue_download_concurrent_threads,
                batch_size=jira_issues_batch_size,
                expand_fields=["changelog"],
                include_fields=include_fields,
                exclude_fields=exclude_fields,
                pull_from=jira_config.pull_from,  # NOTE: We never pull any parents that were last updated before our pull from date! Jira Load WILL NOT load these
                use_jql_enhanced_search=use_jql_enhanced_search,
            )
            with record_span('write_jira_parent_issues_batch_to_local_or_s3'):
                parents_batch_result: IssueDownloadingResult = load_issues_in_batches(
                    issues_to_download=issues_by_ids_generator,
                    ingest_io_helper=ingest_io_helper,
                    ingest_config=ingest_config,
                    jira_config=ingest_config.jira_config,
                    batch_number_start=file_batch_cursor,
                )
                issue_ids_not_uploaded.update(parents_batch_result.issue_ids_too_large_to_upload)

            all_known_ids.update(parents_batch_result.downloaded_ids)
            all_downloaded_ids.update(parents_batch_result.downloaded_ids)
            users_found_across_all_issues.update(parents_batch_result.users_found)
            more_parent_ids = parents_batch_result.discovered_parent_ids - all_known_ids

            if jira_config.recursively_download_parents:
                # TODO: We could probably reduce the numbers of files we upload by grouping
                # all extra parents in joined files, but I don't think it's a big deal to upload
                # a large number of small files
                file_batch_cursor += parents_batch_result.total_batches
                to_be_downloaded_parent_ids = more_parent_ids
                depth_level += 1
            else:
                logging_helper.send_to_agent_log_file(
                    f'recursively_download_parents is set to {jira_config.recursively_download_parents}. Exiting execution'
                )
                break

    # Write Issue IDs that we've downloaded
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraIssuesIdsDownloaded.value,
        json_data=[int(issue_id) for issue_id in all_downloaded_ids],
        subdirectory=SubDirectory.JIRA,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
    )

    # Write Issue IDs that we skipped uploading
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraIssuesIdsSkipped.value,
        json_data={'skipped_issue_ids': [int(issue_id) for issue_id in issue_ids_not_uploaded]},
        subdirectory=SubDirectory.JIRA,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
    )

    add_telemetry_fields({'jira_downloaded_issue_count': len(all_downloaded_ids)})
    add_telemetry_fields(
        {
            'users_found_across_all_issues': len(users_found_across_all_issues),
            # Also log what users are in our DB, to make more information available
            # in our telemetry data
            'users_in_jellyfish_db': len(jira_config.user_keys_in_jellyfish),
        }
    )
    return all_downloaded_ids, ids_to_delete, users_found_across_all_issues


@jelly_trace
@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def load_and_push_jira_to_s3(ingest_config: IngestionConfig) -> bool:
    """Loads data from JIRA, Dumps it to disk, and then uploads that data to S3

    All configuration for this object is done via the JIRAIngestionConfig object

    Args:
        ingest_config (IngestionConfig): A dataclass that holds several different configuration args for this task

    Returns:
        bool: Returns True on Success
    """
    init_jf_ingest_run(ingestion_config=ingest_config)
    company_slug = ingest_config.company_slug
    add_telemetry_fields({'company_slug': company_slug})
    return _run_load_and_push_to_s3(ingest_config=ingest_config)


def _run_load_and_push_to_s3(ingest_config: IngestionConfig) -> bool:
    """Inner function for load_and_push_jira_to_s3

    Args:
        ingest_config (IngestionConfig): A dataclass that holds several different configuration args for this task

    Returns:
        bool: Returns True on Success
    """
    logger.info('Beginning load_and_push_jira_to_s3')
    if not ingest_config.save_locally and not ingest_config.upload_to_s3:
        logger.error(
            f'Configuration error! Ingestion configuration must have either save_locally or upload_to_s3 set to True!'
        )
        raise Exception(
            'Save Locally and Upload to S3 are both set to False! Set one to true or no data will be saved!'
        )

    if not (jira_config := ingest_config.jira_config):
        raise Exception(
            f'load_and_push_to_s3 (Jira Download) was provided without a valid Jira Config!'
        )

    if jira_config.skip_issues and jira_config.only_issues:
        raise Exception(
            'skip_issues and only_issues have both been set to True! This is an invalid JF Ingest object'
        )

    logging_helper.send_to_agent_log_file(f"Feature flags: {jira_config.feature_flags}")

    logger.info("Using local version of ingest")

    #######################################################################
    # SET UP JIRA CONNECTIONS (Basic and Potentially Atlassian Connect)
    # First create test connection with v3 to check JQL Enhanced Search availability
    #######################################################################

    # Test if JQL Enhanced Search is available
    use_jql_enhanced_search = is_jql_enhanced_search_available(
        jira_config=jira_config,
        jql_enhanced_search_enabled=jira_config.feature_flags.get(
            Constants.JQL_ENHANCED_SEARCH_ENABLED, False
        ),
        force_legacy_api=jira_config.feature_flags.get(Constants.FORCE_LEGACY_API, False),
    )

    # Create final connections with appropriate API version
    jira_basic_connection = get_jira_connection(
        config=jira_config,
        auth_method=JiraAuthMethod.BasicAuth,
        use_jql_enhanced_search=use_jql_enhanced_search,
    )

    jira_atlas_connect_connection = (
        get_jira_connection(
            config=jira_config,
            auth_method=JiraAuthMethod.AtlassianConnect,
            use_jql_enhanced_search=use_jql_enhanced_search,
        )
        if JiraAuthMethod.AtlassianConnect in jira_config.available_auth_methods
        else None
    )
    # There is an ongoing effort to cut all things over to Atlassian Connect only,
    # but it is a piecewise migration for now.
    # OJ-29745
    jira_connect_or_fallback_connection = jira_basic_connection
    using_connect_as_primary_auth = False

    if jira_config.feature_flags.get("lusca-auth-always-use-connect-for-atlassian-apis-Q423"):
        logging_helper.send_to_agent_log_file("Will use connect for most API calls")

        if not isinstance(jira_atlas_connect_connection, JIRA):
            raise AtlassianConnectException(
                'Atlassian Connect Connection is not a JIRA object, unable to proceed'
            )

        jira_connect_or_fallback_connection = jira_atlas_connect_connection
        using_connect_as_primary_auth = True
    else:
        logging_helper.send_to_agent_log_file("Will use basic auth for most API calls")

    #######################################################################
    # Init IO Helper
    #######################################################################
    ingest_io_helper = IngestIOHelper(ingest_config=ingest_config)

    S3_OR_JELLYFISH_LOG_STATEMENT = (
        's3' if ingest_config.ingest_type == IngestionType.DIRECT_CONNECT else 'jellyfish'
    )
    if ingest_config.save_locally:
        logger.info(f"Data will be saved locally to: {ingest_io_helper.local_file_path}")
    else:
        logger.info("Data will not be saved locally")
    if ingest_config.upload_to_s3:
        logger.info(f"Data will be submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")
    else:
        logger.info(f"Data will NOT be submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")

    ########################################################################
    # Normalize Jira Config to the Timezone that this Jira Instance exists
    # within
    ########################################################################
    instance_metadata = jira_connect_or_fallback_connection.myself()
    timezone_locale = instance_metadata.get('timeZone', '')
    logger.info(
        f'Jira Instance Timezone Locale detected as: {timezone_locale}. Datetimes will be normalized to this timezone locale'
    )
    try:
        jira_config.normalize_datetimes_to_timezone_locale(timezone_locale)
    except Exception as e:
        # No error should get raised, but just in case
        logger.error(
            f'Failed to normalize datetimes to timezone locale: {timezone_locale}. Error: {e}'
        )

    #######################################################################
    # Jira Projects
    #######################################################################
    projects_and_versions = download_projects_and_versions_and_components(
        jira_connection=jira_connect_or_fallback_connection,
        is_agent_run=ingest_config.ingest_type == IngestionType.AGENT,
        jellyfish_project_ids_to_keys=jira_config.jellyfish_project_ids_to_keys,
        jellyfish_issue_metadata=jira_config.jellyfish_issue_metadata,
        include_projects=jira_config.include_projects,
        exclude_projects=jira_config.exclude_projects,
        include_categories=jira_config.include_project_categories,
        exclude_categories=jira_config.exclude_project_categories,
    )

    project_ids = {proj["id"] for proj in projects_and_versions}
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraProjectsAndVersions.value,
        json_data=projects_and_versions,
        subdirectory=SubDirectory.JIRA,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
    )

    #######################################################################
    # Optionally Fetch Global Components
    #######################################################################
    if jira_config.download_global_components:
        global_components = download_global_components(
            jira_connection=jira_connect_or_fallback_connection,
        )
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraGlobalComponents.value,
            subdirectory=SubDirectory.JIRA,
            json_data=global_components,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )
    else:
        logger.info('Skipping download of global components')

    #######################################################################
    # Jira Fields
    # NOTE: Here we need to download the fields from the fields endpoint
    # and then filter out fields based on the include/exclude list.
    # The tricky part here is that it's possible for customers to provide
    # us either a fields Unique Key (something like customfield_10008), or
    # the fields human readable name (something like "Sprint"). We must
    # support BOTH use cases!
    # To do so, grab all fields data from the API endpoint. Then, expand
    # and normalize the field exclude/include lists to contain both the
    # ID and the name. THEN filter out fields by their ID
    # Once fields have been normalized and expanded we must overwrite
    # the jira_config object to have the expanded list. That way we can
    # pass both the IDs and Keys to our issue downloading logic, which needs
    # both for IDs for grabbing fields AND names for filtering the changelog
    #######################################################################
    jira_fields_data = download_fields(
        jira_connect_or_fallback_connection,
    )
    normalized_include_fields = expand_and_normalize_jira_fields(
        jira_fields_data, jira_config.include_fields
    )
    normalized_exclude_fields = expand_and_normalize_jira_fields(
        jira_fields_data, jira_config.exclude_fields
    )

    users: list[dict] = []
    if not jira_config.only_issues:

        filters = _construct_field_filters(
            include_field_ids=[f.jira_field_id for f in normalized_include_fields],
            exclude_field_ids=[f.jira_field_id for f in normalized_exclude_fields],
        )

        filtered_jira_fields = [
            field for field in jira_fields_data if all(filt(field) for filt in filters)
        ]

        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraFields.value,
            subdirectory=SubDirectory.JIRA,
            json_data=filtered_jira_fields,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        ######################################################################
        # Jira Users
        #######################################################################
        if not jira_config.skip_downloading_users or not jira_config.user_keys_in_jellyfish:
            users = download_users(
                jira_basic_connection=jira_connect_or_fallback_connection,
                jira_atlas_connect_connection=(
                    jira_atlas_connect_connection if jira_config.should_augment_emails else None
                ),  # Use AtlasConnect for 'augment with email' subtask
                gdpr_active=jira_config.gdpr_active,
                search_users_by_letter_email_domain=jira_config.search_users_by_letter_email_domain,
                required_email_domains=jira_config.required_email_domains,
                is_email_required=jira_config.is_email_required,
                using_connect_as_primary_auth=using_connect_as_primary_auth,
            )
        # Upload users, even if there aren't any
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraUsers.value,
            subdirectory=SubDirectory.JIRA,
            json_data=users,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Resolutions
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraResolutions.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_resolutions(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Issue Types
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssueTypes.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_issuetypes(
                jira_connect_or_fallback_connection, project_ids=list(project_ids)
            ),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Link Types
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraLinkTypes.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_issuelinktypes(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Priorities
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraPriorities.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_priorities(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Statuses
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraStatuses.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_statuses(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Boards, Sprints, and Links
        #######################################################################
        if jira_config.download_boards:
            boards, sprints, links = download_boards_and_sprints(
                jira_connect_or_fallback_connection,
                jira_config.download_sprints,
                jira_config.feature_flags.get(Constants.PULL_SPRINTS_BY_ID, False),
                jira_config.feature_flags.get(Constants.PULL_SPRINTS_BY_BOARD_MAX_WORKERS, 5),
                jira_config.feature_flags.get(Constants.PULL_SPRINTS_BY_ID_MAX_WORKERS, 5),
                jira_config.feature_flags.get(Constants.PULL_SPRINTS_SKIP_INACTIVE, False),
                filter_boards_by_projects=(
                    set(project_ids) if jira_config.filter_boards_by_projects else None
                ),
            )
        else:
            boards, sprints, links = [], [], []
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraBoards.value,
            subdirectory=SubDirectory.JIRA,
            json_data=boards,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraSprints.value,
            subdirectory=SubDirectory.JIRA,
            json_data=sprints,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraBoardSprintLinks.value,
            subdirectory=SubDirectory.JIRA,
            json_data=links,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

    if not jira_config.skip_issues:
        api_type = "JQL Enhanced Search" if use_jql_enhanced_search else "Legacy"
        logger.info(f"Using {api_type} API for Jira operations")

        # For Jira Issue operations we need to determine the batch size that
        # the JIRA provider will limit us to
        jira_issues_batch_size = get_jira_search_batch_size(
            jira_connection=jira_connect_or_fallback_connection,
            optimistic_batch_size=jira_config.issue_batch_size,
            use_jql_enhanced_search=use_jql_enhanced_search,
        )

        all_downloaded_issue_ids, all_deleted_issue_ids, users_found_across_all_issues = (
            pull_issues_wrapper(
                jira_connect_or_fallback_connection,
                jira_config=jira_config,
                projects_and_versions=projects_and_versions,
                include_fields=normalized_include_fields,
                exclude_fields=normalized_exclude_fields,
                jira_issues_batch_size=jira_issues_batch_size,
                ingest_io_helper=ingest_io_helper,
                ingest_config=ingest_config,
                use_jql_enhanced_search=use_jql_enhanced_search,
            )
        )

        logger.info(f'{len(all_deleted_issue_ids)} issues have been detected as being deleted')
        # Write issues that got deleted
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssuesIdsDeleted.value,
            subdirectory=SubDirectory.JIRA,
            json_data=list(all_deleted_issue_ids),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        # For now, don't download any user data. Instead, upload the user data
        # so that we can see how many/which users would get downloaded
        logger.info(
            f'Found {len(users_found_across_all_issues)} unique users across all issues (and we have already pulled {len(users)} users in bulk)'
        )
        if jira_config.feature_flags.get(Constants.CHECK_ISSUES_FOR_EXTRA_USERS_AND_PULL):
            _process_additional_users_from_issues(
                jira_basic_connection=jira_basic_connection,
                jira_atlas_connect_connection=jira_atlas_connect_connection,
                ingest_config=ingest_config,
                jira_config=jira_config,
                ingest_io_helper=ingest_io_helper,
                users_downloaded_in_bulk=users,
                user_metadata_from_issues=users_found_across_all_issues,
            )

        #######################################################################
        # Jira Work Logs
        #######################################################################
        if jira_config.download_worklogs:
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraWorklogs.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_worklogs(
                    jira_connect_or_fallback_connection,
                    all_downloaded_issue_ids,
                    jira_config.work_logs_pull_from,
                ),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )
    else:
        logger.info(
            f"Skipping issues and worklogs bc config.skip_issues is {jira_config.skip_issues}"
        )

    if ingest_config.save_locally:
        logger.info(f"Data has been saved locally to: {ingest_io_helper.local_file_path}")
    else:
        logger.info(
            f"Data has not been saved locally, because save_locally was set to false in the ingest config!"
        )

    if ingest_config.upload_to_s3:
        logger.info(f"Data has been submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")
    else:
        logger.info(f"Data was not submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")

    download_metadata = {
        'company_slug': ingest_config.company_slug,
        'ingest_type': ingest_config.ingest_type.value if ingest_config.ingest_type else None,
        'users_downloaded': not jira_config.skip_downloading_users,
        'boards_downloaded': jira_config.download_boards,
        'sprints_downloaded': jira_config.download_boards and jira_config.download_sprints,
        'issues_downloaded': not jira_config.skip_issues,
    }

    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraDownloadMetadata.value,
        subdirectory=SubDirectory.JIRA,
        json_data=download_metadata,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
    )

    return True


def _construct_field_filters(include_field_ids: list[str], exclude_field_ids: list[str]) -> list:
    """Filter out fields we want to exclude, provided exclude and include lists must have the Jira
    field ID names in them

    Args:
        include_field_ids (list[str]): A list of fields (by their Jira Field ID value)
        exclude_field_ids (list[str]): A list of fields (by their Jira Field ID value)

    Returns:
        list: A list of filter functions to filter out fields we don't want to submit to Jellyfish
    """
    filters = []
    if include_field_ids:
        filters.append(lambda field: field["id"] in include_field_ids)
    if exclude_field_ids:
        filters.append(lambda field: field["id"] not in exclude_field_ids)

    return filters


def _process_additional_users_from_issues(
    jira_basic_connection: JIRA,
    jira_atlas_connect_connection: Optional[JIRA],
    ingest_config: IngestionConfig,
    jira_config: JiraDownloadConfig,
    ingest_io_helper: IngestIOHelper,
    users_downloaded_in_bulk: list[dict],
    user_metadata_from_issues: list[UserMetadata],
):
    user_key = get_user_key(jira_config.gdpr_active)
    downloaded_user_keys = set([user[user_key] for user in users_downloaded_in_bulk])
    all_known_user_keys = downloaded_user_keys.union(jira_config.user_keys_in_jellyfish)
    user_data_to_download: list[UserMetadata] = []
    for user in user_metadata_from_issues:
        if user.user_key not in all_known_user_keys:
            user_data_to_download.append(user)

    additional_users = download_users_by_urls(
        jira_basic_connection=jira_basic_connection,
        jira_atlas_connect_connection=jira_atlas_connect_connection,
        user_urls=[u.self_url for u in user_data_to_download],
        required_email_domains=jira_config.required_email_domains,
        is_email_required=jira_config.is_email_required,
        num_parallel_threads=jira_config.issue_download_concurrent_threads,
    )

    all_users = []
    all_users.extend(users_downloaded_in_bulk)
    all_users.extend(additional_users)
    logger.info(
        f'Found an additional {len(additional_users)} users to the already downloaded {len(users_downloaded_in_bulk)} users downloaded in bulk. Uploading these users now'
    )
    # Overwrite users that were potentially already uploaded
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraUsers.value,
        subdirectory=SubDirectory.JIRA,
        json_data=all_users,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
    )



from typing import Optional


class GitProviderUnavailable(Exception):
    pass


class GqlRateLimitExceededException(Exception):
    pass


class GqlPageTimeoutException(Exception):
    pass


class GitAuthenticationException(Exception):
    def __init__(self, *args, original_exception: Optional[Exception] = None):
        self.original_exception = original_exception
        super().__init__(*args)


class GitAuthorizationException(Exception):
    pass



from unittest.mock import patch

import pytest
from jira import JIRAError

from jf_ingest.config import JiraAuthConfig
from jf_ingest.jf_jira.auth import (
    JiraAuthenticationException,
    JiraAuthMethod,
    get_jira_connection,
)
from jf_ingest.jf_jira.exceptions import JiraRetryLimitExceeded


def _get_jira_auth_config(args):
    company_slug = "test_company"
    url = "https://test-co.atlassian.net/"
    return JiraAuthConfig(company_slug=company_slug, url=url, gdpr_active=True, **args)


def test_jira_conn_basic_auth_no_auth():
    config = _get_jira_auth_config({"available_auth_methods": [JiraAuthMethod.BasicAuth]})

    with pytest.raises(RuntimeError) as excinfo:
        get_jira_connection(config)

    assert (
        str(excinfo.value)
        == f"No valid basic authentication mechanism for {config.url} - need a username/password combo or a personal access token"
    )


def test_jira_conn_basic_auth_user_password():
    config = _get_jira_auth_config(
        {
            "available_auth_methods": [JiraAuthMethod.BasicAuth],
            "user": "jira_user@test.co",
            "password": "test_password",  # pragma: allowlist secret
        }
    )

    # TEST BASIC AUTH ERROR LOGIC
    with patch("jf_ingest.jf_jira.auth.JIRA", side_effect=JIRAError(status_code=401)):
        with pytest.raises(JiraAuthenticationException):
            get_jira_connection(config, max_retries=0)

    # TEST BASIC AUTH ERROR LOGIC
    with patch("jf_ingest.jf_jira.auth.JIRA", side_effect=JIRAError(status_code=403)):
        with pytest.raises(JiraAuthenticationException):
            get_jira_connection(config, max_retries=0)


def test_jira_conn_basic_auth_personal_access_token():
    config = _get_jira_auth_config(
        {
            "available_auth_methods": [JiraAuthMethod.BasicAuth],
            "personal_access_token": "a_PAT",
        }
    )

    # TEST BASIC AUTH ERROR LOGIC
    with patch("jf_ingest.jf_jira.auth.JIRA", side_effect=JIRAError(status_code=401)):
        with pytest.raises(JiraAuthenticationException):
            get_jira_connection(config, max_retries=0)

    # TEST BASIC AUTH ERROR LOGIC
    with patch("jf_ingest.jf_jira.auth.JIRA", side_effect=JIRAError(status_code=403)):
        with pytest.raises(JiraAuthenticationException):
            get_jira_connection(config, max_retries=0)


def test_jira_conn_rest_api_version_selection():
    """Test that the correct REST API version is set based on use_jql_enhanced_search parameter"""
    config = _get_jira_auth_config(
        {
            "available_auth_methods": [JiraAuthMethod.BasicAuth],
            "personal_access_token": "a_PAT",
        }
    )

    # Mock JIRA to capture the connection kwargs
    with patch("jf_ingest.jf_jira.auth.JIRA") as mock_jira:
        # Test default behavior (use_jql_enhanced_search=False) - should not set REST API version (uses default v2)
        get_jira_connection(config, max_retries=0)
        
        # Verify JIRA was called without explicit REST API version (uses default v2)
        call_kwargs = mock_jira.call_args[1]
        assert "rest_api_version" not in call_kwargs["options"]

    with patch("jf_ingest.jf_jira.auth.JIRA") as mock_jira:
        # Test JQL Enhanced Search (use_jql_enhanced_search=True) - should use REST API v3
        get_jira_connection(config, max_retries=0, use_jql_enhanced_search=True)
        
        # Verify JIRA was called with REST API version 3
        call_kwargs = mock_jira.call_args[1]
        assert call_kwargs["options"]["rest_api_version"] == "3"



import logging
from datetime import datetime, timedelta, timezone
from itertools import chain
from typing import Any, Dict, Generator, List, Optional, cast

from dateutil import parser

from jf_ingest import logging_helper
from jf_ingest.config import GitConfig
from jf_ingest.jf_git.adapters import GitAdapter
from jf_ingest.jf_git.clients.azure_devops import AzureDevopsClient
from jf_ingest.jf_git.standardized_models import (
    PullRequestReviewState,
    StandardizedBranch,
    StandardizedCommit,
    StandardizedFileData,
    StandardizedJFAPIPullRequest,
    StandardizedLabel,
    StandardizedOrganization,
    StandardizedPullRequest,
    StandardizedPullRequestAuthor,
    StandardizedPullRequestComment,
    StandardizedPullRequestMetadata,
    StandardizedPullRequestReview,
    StandardizedPullRequestReviewAuthor,
    StandardizedRepository,
    StandardizedTeam,
    StandardizedUser,
)
from jf_ingest.utils import hash_filename


class AzureDevopsAdapter(GitAdapter):
    def __init__(self, config: GitConfig):
        # Git Config options
        self.client = AzureDevopsClient(config.git_auth_config)
        self.config = config
        self.repo_id_to_name_lookup: Dict[str, str] = {}
        self.repo_id_to_project_name: Dict[str, str] = {}
        self.org_to_project_names: Dict[str, List[str]] = {}

        for org in self.config.git_organizations:
            if len(org.split('/')) == 2:
                org_name, project_name = org.split('/')
                if org_name in self.org_to_project_names:
                    self.org_to_project_names[org_name].append(project_name)
                else:
                    self.org_to_project_names[org_name] = [project_name]
            else:
                self.org_to_project_names[org] = []

    @property
    def supports_get_all_users(self) -> bool:
        """
        mpk 5/12/23: Azure DevOps Server 2020 (on-prem, encountered for at least one client) does NOT support a
        get_all_users method in the API. Cloud-based Azure DevOps Services (i.e., dev.azure.com) does. We
        haven't yet encountered other ADO server versions (like ADO Server 2022), so don't know whether or
        not those other versions will support get_all_users. For now I'll just assume that anything that
        runs the API version "6.1*" (like ADO Server 2020) will NOT support get_all_users.

        Returns:
            bool: True if we are in a version that is not 6.1
        """
        return not self.client.api_version.startswith('6.1')

    def _project_name_from_repo(self, standardized_repo: StandardizedRepository) -> str:
        '''
        The Project Name should be pre-loaded into the self.repo_id_to_project_name object,
        which gets set in the adapter get_repos function, and mapped to the repo ID. If
        it isn't, call the get_all_repos() client function again and save the results
        to the adapter.

        Returns the project name of the provided standardized_repo
        '''
        if project_name := self.repo_id_to_project_name.get(standardized_repo.id):
            return project_name
        else:
            logging_helper.send_to_agent_log_file(
                f'Repos were not preloaded? Loading them now', level=logging.WARNING
            )
            for repo in self.client.get_all_repos(org_name=standardized_repo.organization.login):
                self.repo_id_to_project_name[repo['id']] = repo['project']['name']

        return self.repo_id_to_project_name[standardized_repo.id]

    def get_organizations(self) -> List[StandardizedOrganization]:
        return [
            _standardize_organization(org, self.client.base_url)
            for org in self.org_to_project_names.keys()  # Set up the login to match only the Org, not the included project name(s)
        ]

    def get_api_scopes(self) -> str:
        raise NotImplementedError()

    def get_users(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedUser, None, None]:
        i = 1
        for api_user in self.client.get_graph_users(org_name=standardized_organization.login):
            standardized_user = _standardize_graph_user(api_user)
            if standardized_user:
                yield standardized_user
                if limit and i >= limit:
                    return
                i += 1

    def get_teams(
        self, standardized_organization: StandardizedOrganization, limit: Optional[int] = None
    ) -> Generator[StandardizedTeam, None, None]:
        for i, api_team in enumerate(
            self.client.get_teams(org_name=standardized_organization.login), start=1
        ):
            api_team_members = self.client.get_team_users(
                org_name=standardized_organization.login, team_descriptor=api_team['descriptor']
            )
            yield _standardize_team(api_team, api_team_members=api_team_members)
            if limit and i >= limit:
                return

    def get_repos(
        self,
        standardized_organization: StandardizedOrganization,
        limit: Optional[int] = None,
        only_private: bool = False,
    ) -> Generator[StandardizedRepository, None, None]:
        # Note: the only_private param is unused here, since we don't use the 'include hidden repositories'
        # flag when querying for repos from the ADO API. If this changes in the future, we'll need to add
        # this functionality
        org_name = standardized_organization.login
        project_names_in_org = self.org_to_project_names.get(org_name, [])

        if project_names_in_org:
            org_project_names = [
                f'{org_name}/{_project_name}' for _project_name in project_names_in_org
            ]
        else:
            org_project_names = [org_name]

        for org_project_name in org_project_names:
            for idx, api_repo in enumerate(
                self.client.get_all_repos(org_name=org_project_name), start=1
            ):
                project_name = api_repo['project']['name']
                if f'{org_name}/{project_name}' in self.config.excluded_organizations:
                    continue
                self.repo_id_to_project_name[api_repo['id']] = project_name
                yield _standardize_repo(
                    org=standardized_organization,
                    api_repo=api_repo,
                )

                if limit and idx >= limit:
                    return

    def get_repos_count(
        self, standardized_organization: StandardizedOrganization, only_private: bool = False
    ) -> int:
        """
        Get the count of repos in the organization. only_private is ignored here, since we don't normally pull
        hidden ADO repos

        Args:
            standardized_organization: A StandardizedOrganization object
            only_private: Ignored for ADO

        Returns:
            int: The count of repos in the organization
        """
        org_name = standardized_organization.login
        project_names_in_org = self.org_to_project_names.get(org_name, [])

        if project_names_in_org:
            org_project_names = [
                f'{org_name}/{_project_name}' for _project_name in project_names_in_org
            ]
        else:
            org_project_names = [org_name]

        count = 0
        for org_project_name in org_project_names:
            count += len(self.client.get_all_repos(org_name=org_project_name))

        return count

    def get_commits_for_default_branch(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        for i, api_commit in enumerate(
            self.client.get_commits(
                org_name=standardized_repo.organization.login,
                project_name=self._project_name_from_repo(standardized_repo),
                repo_id=standardized_repo.id,
                branch_name=standardized_repo.default_branch_name,
                from_date=pull_since,
                to_date=pull_until,
            ),
            start=1,
        ):
            yield _standardize_commit(
                api_commit,
                branch_name=standardized_repo.default_branch_name,
                standardized_repo=standardized_repo,
            )
            if limit and i >= limit:
                return

    def get_branches_for_repo(
        self,
        standardized_repo: StandardizedRepository,
        pull_branches: Optional[bool] = False,
        limit: Optional[int] = None,
    ) -> Generator[StandardizedBranch, None, None]:
        if pull_branches:
            for idx, api_branch in enumerate(
                self.client.get_branches(
                    org_name=standardized_repo.organization.login,
                    project_name=self._project_name_from_repo(standardized_repo),
                    repo_id=standardized_repo.id,
                ),
                start=1,
            ):
                if standardized_branch := _standardize_branch(
                    api_branch=api_branch, standardized_repo=standardized_repo
                ):
                    yield standardized_branch

                if limit and idx >= limit:
                    return
        else:
            # Above, if we're pulling all branches, it's safe to assume that the default branch
            # will be included in that
            # When we don't pull all branches, always return default branch
            if standardized_repo.default_branch_name:
                yield StandardizedBranch(
                    repo_id=standardized_repo.id,
                    name=standardized_repo.default_branch_name,
                    sha=standardized_repo.default_branch_sha,
                    is_default=True,
                )

    def get_commits_for_branches(
        self,
        standardized_repo: StandardizedRepository,
        branches: List[StandardizedBranch],
        pull_since: Optional[datetime] = None,
        pull_until: Optional[datetime] = None,
        limit: Optional[int] = None,
    ) -> Generator[StandardizedCommit, None, None]:
        pulled_commit_count = 0

        for branch_name in self.get_filtered_branches(standardized_repo, branches):
            try:
                login = standardized_repo.organization.login
                for api_commit in self.client.get_commits(
                    org_name=login,
                    project_name=self._project_name_from_repo(standardized_repo),
                    repo_id=standardized_repo.id,
                    branch_name=branch_name,
                    from_date=pull_since,
                    to_date=pull_until,
                ):
                    yield _standardize_commit(
                        api_commit=api_commit,
                        branch_name=branch_name,
                        standardized_repo=standardized_repo,
                    )

                    pulled_commit_count += 1
                    if limit and pulled_commit_count >= limit:
                        return

            except Exception as e:
                logging_helper.send_to_agent_log_file(
                    f'Got exception for branch {branch_name}: {e}. Skipping...',
                    level=logging.WARNING,
                )

    def get_pr_updated_date(
        self, standardized_repo: StandardizedRepository, api_pr: Dict
    ) -> datetime:
        if 'closedDate' in api_pr:
            return cast(datetime, parser.parse(api_pr['closedDate']))

        org_name = standardized_repo.organization.login
        project_name = self._project_name_from_repo(standardized_repo)
        repo_id = standardized_repo.id
        pr_iterations = self.client.get_pull_request_iterations(
            org_name=org_name,
            project_name=project_name,
            repo_id=repo_id,
            pr_id=api_pr['pullRequestId'],
        )
        assert len(pr_iterations) > 0
        # Get the latest iteration and grab the updatedDate value
        return cast(datetime, parser.parse(pr_iterations[-1]['updatedDate']))

    def get_pr_metadata(
        self,
        standardized_repo: StandardizedRepository,
        limit: Optional[int] = None,
        pr_pull_from_date: Optional[datetime] = None,
    ) -> Generator[StandardizedPullRequestMetadata, None, None]:
        supports_pr_filter_func = self.git_provider_pr_endpoint_supports_date_filtering
        raise NotImplementedError(
            'This function is NOT implemented, because ADO supports PR date filtering '
            f'({supports_pr_filter_func.__name__} is {supports_pr_filter_func()})'
        )

    def git_provider_pr_endpoint_supports_date_filtering(self):
        return True

    def _get_prs_helper(
        self,
        standardized_repo: StandardizedRepository,
        start_window: Optional[datetime],
        end_window: Optional[datetime],
    ) -> Generator[Dict, None, None]:
        tomorrow = datetime.now().astimezone(timezone.utc) + timedelta(days=1)

        def _get_prs_from_client_wrapper(
            _end_window: Optional[datetime], status: str, filter_by: str
        ):
            return self.client.get_pull_requests(
                org_name=standardized_repo.organization.login,
                project_name=self._project_name_from_repo(standardized_repo=standardized_repo),
                repo_id=standardized_repo.id,
                start_window=start_window,
                end_window=_end_window,
                status=status,
                filter_by=filter_by,
            )

        abandoned_prs_generator = _get_prs_from_client_wrapper(end_window, 'abandoned', 'closed')
        completed_prs_generator = _get_prs_from_client_wrapper(end_window, 'completed', 'closed')
        active_prs = _get_prs_from_client_wrapper(tomorrow, 'active', 'open')
        return cast(
            Generator[Dict, None, None],
            chain(abandoned_prs_generator, completed_prs_generator, active_prs),
        )

    def get_prs(
        self,
        standardized_repo: StandardizedRepository,
        pull_files_for_pr: bool = False,
        hash_files_for_prs: bool = False,
        limit: Optional[int] = None,  # Not used by ADO adapter
        start_cursor: Any = 0,
        start_window: Optional[datetime] = None,
        end_window: Optional[datetime] = None,
    ) -> Generator[StandardizedPullRequest, None, None]:
        try:
            login = standardized_repo.organization.login
            repo_id = standardized_repo.id
            project_name = self._project_name_from_repo(standardized_repo)

            api_prs = self._get_prs_helper(
                standardized_repo=standardized_repo,
                start_window=start_window,
                end_window=end_window,
            )

            for i, api_pr in enumerate(
                api_prs,
                start=1,
            ):
                try:
                    pr_id = api_pr['pullRequestId']
                    head_branch = api_pr['sourceRefName'].replace('refs/heads/', '')
                    base_branch = api_pr['targetRefName'].replace('refs/heads/', '')
                    files = {}
                    changed_count = 0
                    additions = 0
                    deletions = 0
                    if 'lastMergeTargetCommit' in api_pr and 'lastMergeCommit' in api_pr:
                        change_counts = self.client.get_pull_request_changes_counts(
                            org_name=login,
                            repo_id=standardized_repo.id,
                            base_sha=api_pr['lastMergeTargetCommit']['commitId'],
                            target_sha=api_pr['lastMergeCommit']['commitId'],
                        )
                        changed_count = sum(change_counts.values())
                        additions = change_counts.get('Add', 0)
                        deletions = change_counts.get('Delete', 0)
                        if pull_files_for_pr:
                            changes = self.client.get_pull_request_diff(
                                org_name=login,
                                repo_id=standardized_repo.id,
                                base_sha=api_pr['lastMergeTargetCommit']['commitId'],
                                target_sha=api_pr['lastMergeCommit']['commitId'],
                            )
                            files = {
                                c['item']['path']: StandardizedFileData(
                                    status=c.get('changedType', ''),
                                    changes=0,
                                    additions=0,
                                    deletions=0,
                                )
                                for c in changes
                                if not c['item'].get('isFolder', False)
                            }
                            if hash_files_for_prs:
                                files = {
                                    hash_filename(file_path): file_data
                                    for file_path, file_data in files.items()
                                }
                        else:
                            files = {}

                    pr_updated_at = self.get_pr_updated_date(
                        standardized_repo=standardized_repo, api_pr=api_pr
                    )

                    commits = [
                        _standardize_commit(
                            api_commit=api_commit,
                            branch_name=head_branch,
                            standardized_repo=standardized_repo,
                        )
                        for api_commit in self.client.get_pr_commits(
                            org_name=login,
                            project_name=project_name,
                            repo_id=repo_id,
                            pr_id=pr_id,
                        )
                    ]
                    # If a merge commit is in here, retroactively mark one of the commits
                    # as a merge commit
                    if api_pr['status'] == 'completed' and 'lastMergeCommit' in api_pr:
                        for commit in commits:
                            if commit.hash == api_pr['lastMergeCommit']['commitId']:
                                commit.is_merge = True

                    comments: List[StandardizedPullRequestComment] = []
                    reviews: List[StandardizedPullRequestReview] = []
                    for comment_thread in self.client.get_pr_comment_threads(
                        org_name=login, project_name=project_name, repo_id=repo_id, pr_id=pr_id
                    ):
                        comment_thread_id = comment_thread['id']
                        for api_comment in comment_thread['comments']:
                            comments.append(_standardize_pull_request_comment(api_comment))
                            pr_review = _standardize_pull_request_review(
                                api_comment=api_comment,
                                pr_id=pr_id,
                                comment_thread_id=comment_thread_id,
                            )

                            if pr_review:
                                reviews.append(pr_review)

                    labels = [label['name'] for label in api_pr.get('labels', [])]

                    yield _standardize_pull_request(
                        api_pr,
                        standardized_repo,
                        pr_updated_at=pr_updated_at,
                        commits=commits,
                        comments=comments,
                        approvals=reviews,
                        labels=labels,
                        additions=additions,
                        deletions=deletions,
                        changed_count=changed_count,
                        files=files,
                    )
                    if limit and i >= limit:
                        return
                except Exception:
                    # if something goes wrong with normalizing one of the prs - don't stop pulling. try
                    # the next one.
                    pr_id = api_pr.get('id', '?')
                    logging_helper.send_to_agent_log_file(
                        f'normalizing PR {pr_id} from repo {standardized_repo.name} ({standardized_repo.id}). Skipping...',
                        level=logging.WARNING,
                        exc_info=True,
                    )

        except Exception:
            logging_helper.send_to_agent_log_file(
                f'normalizing PRs from repo {standardized_repo.name} ({standardized_repo.id}). Skipping...',
                level=logging.WARNING,
                exc_info=True,
            )

    def get_pr_authors(
        self, pr_list: list[StandardizedJFAPIPullRequest]
    ) -> list[StandardizedPullRequestAuthor]:
        """
        Get the authors of a list of PRs

        This is currently NOT IMPLEMENTED for ADO instances
        """
        raise NotImplementedError()

    def get_pr_review_authors(
        self, pr_list: list[StandardizedJFAPIPullRequest]
    ) -> list[StandardizedPullRequestReviewAuthor]:
        """
        Get the review authors of a list of PRs

        This is currently NOT IMPLEMENTED for ADO instances
        """
        raise NotImplementedError()


'''

Massage Functions (standardize raw JSON data to StandardizedPython Models)

'''


def _standardize_graph_user(api_user: Dict) -> Optional[StandardizedUser]:
    """Standardize the raw GraphUser type to our StandardizedUser.
    This is for users returned from the "Graph" API.
    https://learn.microsoft.com/en-us/rest/api/azure/devops/graph/users/list?view=azure-devops-rest-7.1&tabs=HTTP#graphuser

    Args:
        api_user (Dict): A raw dictionary from the ADO API representing a GraphUser

    Returns:
        StandardizedUser: A standardized User object
    """
    if not (user_id := api_user.get('descriptor')):
        return None
    return StandardizedUser(
        id=user_id,
        login=api_user['principalName'],
        name=api_user['displayName'],
        email=api_user['mailAddress'],
        url=api_user['url'],
    )


def _standardize_team(api_team: Dict, api_team_members: List[Dict]) -> StandardizedTeam:
    members = []
    for member in api_team_members:
        if standardized_member := _standardize_graph_user(member):
            members.append(standardized_member)

    return StandardizedTeam(
        id=api_team['descriptor'],
        slug=api_team['descriptor'],
        name=api_team['displayName'],
        description=api_team['description'],
        members=members,
    )


def _standardize_repo(org: StandardizedOrganization, api_repo: dict) -> StandardizedRepository:
    # Sanitize Repo Name
    repo_name = api_repo['name']

    # Sanitize URL
    url = api_repo['webUrl']

    # Sanitize Branch Name
    if 'defaultBranch' in api_repo:
        default_branch_name = api_repo['defaultBranch'].replace('refs/heads/', '')
    else:
        default_branch_name = None

    return StandardizedRepository(
        id=api_repo['id'],
        name=repo_name,
        full_name=f'{org.login}/{repo_name}',
        url=url,
        is_fork=api_repo.get('isFork', False),
        organization=org,
        default_branch_name=default_branch_name,
        default_branch_sha=None,  # TODO: looks like this isn't part of the base object, is it important? Do we NEED it?
    )


def _standardize_branch(
    api_branch: dict, standardized_repo: StandardizedRepository
) -> Optional[StandardizedBranch]:
    if not api_branch:
        return None
    if not api_branch.get('name'):
        return None

    branch_name = api_branch['name'].replace('refs/heads/', '')
    return StandardizedBranch(
        repo_id=standardized_repo.id,
        name=api_branch['name'].replace('refs/heads/', ''),
        sha=api_branch['objectId'],
        is_default=branch_name == standardized_repo.default_branch_name,
    )


def _standardize_git_user_data(user: Dict) -> Optional[StandardizedUser]:
    """Standardize the GitUserDate type, which is the type of user returned
    on commits.
    https://learn.microsoft.com/en-us/rest/api/azure/devops/git/commits/get-commits?view=azure-devops-rest-7.1&tabs=HTTP#gituserdate

    Args:
        user (Dict): A GitUserDict returned from the API

    Returns:
        StandardizedUser: The Standardized User
    """
    if not (email := user.get('email')):
        return None
    return StandardizedUser(
        id=email,
        name=user.get('name'),
        login=email,
        email=email,
        url=None,
        account_id=None,
    )


def _standardize_organization(org_login: str, url: str) -> StandardizedOrganization:
    # TODO this is probably wrong
    return StandardizedOrganization(
        id=org_login,
        name=org_login,
        login=org_login,
        url=url,
    )


def _standardize_commit(
    api_commit: dict,
    branch_name: Optional[str],
    standardized_repo: StandardizedRepository,
) -> StandardizedCommit:
    # NOTE: The commit object only returns "GitUserDate" types
    author = _standardize_git_user_data(api_commit['author'])
    return StandardizedCommit(
        hash=api_commit['commitId'],
        url=api_commit['url'],
        message=api_commit['comment'],
        branch_name=branch_name,
        commit_date=parser.parse(api_commit['committer']['date']),
        author_date=parser.parse(api_commit['author']['date']),
        author=author if author else None,
        repo=standardized_repo.short(),
        # XXX: The commit we get back from the API doesn't include the "parents" item; the docs seem
        # to indicate it's supposed to. Without having the "parents" item, we can't tell whether or
        # not the commit is a merge commit. I think we could get the "parents" item if we use the
        # "Get" API to get a single commit rather than the "GetCommits" API to get a list of commits.
        # But that'd cause us to make way more API requests. For now at least let's avoid that and
        # just set is_merge to False (perhaps erroneously) for all commits.
        # NOTE (update): For Pull Requests we WILL mark merge commits properly as part of PR Ingestion.
        # In theory all, if not most, merge commits will come to us via PRs, so we should eventaully
        # mark all merge commits with this strategy
        is_merge=False,
    )


def _standardize_identity_ref_user(user: Dict) -> StandardizedUser:
    """Standardize the ADO IdentityRef User object.
    This is the type commonly returned by the PR related data objects.
    https://learn.microsoft.com/en-us/rest/api/azure/devops/git/pull-request-threads/list?view=azure-devops-rest-7.1&tabs=HTTP#identityref

    Args:
        user (Dict): A raw "IdentityRef" object type from ADO

    Returns:
        StandardizedUser: A standardized User Object
    """
    user_id = user.get('descriptor') or user.get('uniqueName') or user.get('id')
    return StandardizedUser(
        id=str(user_id),
        name=user.get('displayName'),
        login=str(user_id),
        email=user.get('uniqueName'),
        url=user['url'],
        account_id=user['id'],
    )


def _standardize_pull_request_review(
    api_comment: Dict, pr_id: int, comment_thread_id: int
) -> Optional[StandardizedPullRequestReview]:
    """Normalize an API comment into a review.
    State is derived by checking if this is a system comment type, and by
    checking the comment content.
    User is derived from the comment dictionary
    foreign-id is derived from Pull Request ID, Comment Thread ID, and Comment ID. Honestly,
    this probably isn't that useful but at least it's something

    Args:
        api_comment (Dict): Used to derive approval state and commenter
        pr_id (int): Used as part of the Foreign ID field
        comment_thread_id (int): Used as part of the Foreign ID field

    Returns:
        StandardizedPullRequestReview: A standardized Pull Request Review
    """
    api_comment_author = api_comment['author']
    commenter = (
        _standardize_identity_ref_user(api_comment_author)
        if 'uniqueName' in api_comment_author
        else None
    )

    comment_type = api_comment.get('commentType', '')
    comment_content = api_comment.get('content', '')

    # If the comment is an approval, construct a StandardizedPullRequestApproval
    review_state: Optional[PullRequestReviewState] = None
    if comment_type == 'system' and 'voted 5' in comment_content:
        # This is "approved the pull request with suggestions"
        review_state = PullRequestReviewState.APPROVED
    elif comment_type == 'system' and 'voted 10' in comment_content:
        # This is "approved the pull request"
        review_state = PullRequestReviewState.APPROVED
    elif comment_type == 'system' and 'voted -10' in comment_content:
        # This is "rejected the pull request"
        review_state = PullRequestReviewState.DISMISSED
    elif comment_type == 'system' and 'voted -5' in comment_content:
        # This is "is waiting for the author"
        review_state = PullRequestReviewState.CHANGES_REQUESTED
    elif comment_type == 'text':
        review_state = PullRequestReviewState.COMMENTED

    if review_state:
        return StandardizedPullRequestReview(
            user=commenter,
            foreign_id=f"{pr_id}-{comment_thread_id}-{api_comment['id']}",
            review_state=review_state.name,
        )
    return None


def _standardize_pull_request_comment(
    api_comment: Dict,
) -> StandardizedPullRequestComment:
    api_comment_author = api_comment['author']
    commenter = (
        _standardize_identity_ref_user(api_comment_author)
        if 'uniqueName' in api_comment_author
        else None
    )
    comment_body = api_comment.get('content', '')
    return StandardizedPullRequestComment(
        user=commenter,
        body=comment_body,
        created_at=parser.parse(api_comment['publishedDate']),
        system_generated=api_comment.get('commentType') == 'system',
    )


def _standardize_pull_request(
    api_pr: Dict,
    standardized_repo: StandardizedRepository,
    pr_updated_at: datetime,
    commits: List[StandardizedCommit],
    comments: List[StandardizedPullRequestComment],
    approvals: List[StandardizedPullRequestReview],
    labels: List[StandardizedLabel],
    additions: int,
    deletions: int,
    changed_count: int,
    files: Dict[str, StandardizedFileData],
) -> StandardizedPullRequest:
    #
    # PR Status
    pr_status = api_pr['status']
    is_merged = pr_status == 'completed'
    is_closed = pr_status != 'active'
    closed_date = parser.parse(api_pr['closedDate']) if 'closedDate' in api_pr else None
    merge_date = closed_date if is_merged else None

    #
    # User Transformation
    author = _standardize_identity_ref_user(api_pr['createdBy'])

    merge_commits = [c for c in commits if c.is_merge]
    merge_commit = merge_commits[0] if len(merge_commits) == 1 else None
    if merge_commit:
        merged_by = merge_commit.author
    else:
        merged_by = None

    title = api_pr['title']
    body = api_pr.get('description', '')

    base_branch = api_pr['targetRefName'].replace('refs/heads/', '')
    head_branch = api_pr['sourceRefName'].replace('refs/heads/', '')

    return StandardizedPullRequest(
        id=api_pr['pullRequestId'],
        additions=additions,
        deletions=deletions,
        changed_files=changed_count,
        is_closed=is_closed,
        is_merged=is_merged,
        created_at=parser.parse(api_pr['creationDate']),
        updated_at=pr_updated_at,
        merge_date=merge_date,
        closed_date=closed_date,
        title=title,
        body=body,
        url=api_pr['url'],
        base_branch=base_branch,
        head_branch=head_branch,
        author=author,
        merged_by=merged_by if merged_by else None,
        commits=commits,
        merge_commit=merge_commit,
        comments=comments,
        approvals=approvals,
        base_repo=standardized_repo.short(),
        head_repo=standardized_repo.short(),
        labels=labels,
        files=files,
    )
